from typing import Optional, Union

from pydantic import BaseModel

from phidata.infra.k8s.enums.protocol import Protocol


class CreatePort(BaseModel):
    """

    References:
        * https://matthewpalmer.net/kubernetes-app-developer/articles/kubernetes-ports-targetport-nodeport-service.html
    """

    # If specified, this must be an IANA_SVC_NAME and unique within the pod.
    # Each named port in a pod must have a unique name.
    # Name for the port that can be referred to by services.
    name: Optional[str] = None
    # Number of port to expose on the pod's IP address. This must be a valid port number, 0 < x < 65536.
    # This is port the application is running on the container
    container_port: int
    ## If the deployment running this container is exposed by a service
    # The service_port is the port that will be exposed by that service.
    service_port: Optional[int] = None
    # The target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod. usually the same as self.name
    target_port: Optional[Union[str, int]] = None
    # When using a service of type: NodePort or LoadBalancer
    # This is the port on each node on which this service is exposed
    node_port: Optional[int] = None
    protocol: Optional[Protocol] = None
    # host_ip: Optional[str] = None
    # Number of port to expose on the host.
    # If specified, this must be a valid port number, 0 < x < 65536.
    # Most containers do not need this.
    # host_port: Optional[int] = None



from pathlib import Path
from typing import Optional

from pydantic import AnyUrl

from phidata.asset.file import File
from phidata.utils.cli_console import print_info, print_error, print_warning
from phidata.utils.log import logger
from phidata.task import PythonTask, PythonTaskArgs


class DownloadUrlToFileArgs(PythonTaskArgs):
    file: File
    url: str


class DownloadUrlToFile(PythonTask):
    def __init__(
        self,
        file: File,
        url: str,
        name: str = "download_url_to_file",
        task_id: Optional[str] = None,
        dag_id: Optional[str] = None,
        version: Optional[str] = None,
        enabled: bool = True,
    ):
        super().__init__()
        try:
            self.args: DownloadUrlToFileArgs = DownloadUrlToFileArgs(
                file=file,
                url=url,
                name=name,
                task_id=task_id,
                dag_id=dag_id,
                version=version,
                enabled=enabled,
                entrypoint=download_url_to_file,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    @property
    def file(self) -> Optional[File]:
        return self.args.file

    @file.setter
    def file(self, file: File) -> None:
        if file is not None:
            self.args.file = file

    @property
    def url(self) -> str:
        return self.args.url

    @url.setter
    def url(self, url: str) -> None:
        if url is not None:
            self.args.url = url


def download_url_to_file(**kwargs) -> bool:

    import httpx
    import rich.progress

    args = DownloadUrlToFileArgs.from_kwargs(kwargs)
    # logger.info(f"DownloadUrlToFileArgs: {args}")

    file_to_load: Optional[File] = args.file
    if file_to_load is None:
        print_error("File not available")
        return False
    file_path: Optional[Path] = file_to_load.file_path
    if file_path is None:
        print_error("FilePath not available")
        return False
    file_dir: Path = file_path.parent.resolve()

    url = args.url
    if url is None:
        print_error("URL not available")
        return False

    # Download URL contents
    print_info("Downloading")
    print_info(f"  Url: {url}")
    print_info(f"  To : {file_path}")

    # Create the directory if it does not exist
    if not file_dir.exists():
        file_dir.mkdir(parents=True, exist_ok=True)

    bytes_to_download: int = 0
    try:
        with file_path.open("w") as open_file:
            with httpx.stream("GET", url) as response:
                try:
                    headers: httpx.Headers = response.headers
                    # logger.debug("headers: {}".format(headers))
                    # logger.debug("status_code: {}".format(response.status_code))
                    bytes_to_download = int(headers["Content-Length"])
                    # logger.debug("bytes_to_download : {}".format(bytes_to_download))
                except Exception:
                    print_warning(f"Could not get total bytes_to_download from headers")

                with rich.progress.Progress(
                    "[progress.percentage]{task.percentage:>3.0f}%",
                    rich.progress.BarColumn(50),
                    rich.progress.DownloadColumn(),
                    rich.progress.TransferSpeedColumn(),
                ) as progress:
                    download_task = progress.add_task(
                        "Download", total=bytes_to_download
                    )
                    for chunk in response.iter_text():
                        open_file.write(chunk)
                        progress.update(
                            download_task, completed=response.num_bytes_downloaded
                        )
        logger.info("Url downloaded")
        return True
    except Exception as e:
        print_error("Could not download url, please try again")
        print_info("--- stacktrace ---")
        logger.exception(e)
        print_info("--- stacktrace ---")
    return False


def best_guess_file_name(url: AnyUrl, file_name: Optional[str] = None) -> Optional[str]:

    from phidata.utils.dttm import get_today_utc_date_str

    fn: Optional[str] = file_name
    if fn is None:
        _url_path: str = str(url)
        if "/" in _url_path:
            fn = _url_path.split("/")[-1]
    if fn is None:
        fn = get_today_utc_date_str()
    # logger.debug("fn: {}".format(fn))
    return fn



from collections import OrderedDict
from typing import Optional, List

from phidata.app import PhidataApp, PhidataAppArgs
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class ElasticAppArgs(PhidataAppArgs):
    name: str = "elastic"
    version: str = "1"
    enabled: bool = True
    docker_resource_groups: Optional[List[DockerResourceGroup]] = None
    k8s_resource_groups: Optional[List[K8sResourceGroup]] = None
    create_k8s_resource_groups: Optional[List[CreateK8sResourceGroup]] = None


class ElasticApp(PhidataApp):
    def __init__(
        self,
        name: str = "elastic",
        version: str = "1",
        enabled: bool = True,
        docker_resource_groups: Optional[List[DockerResourceGroup]] = None,
        k8s_resource_groups: Optional[List[K8sResourceGroup]] = None,
        create_k8s_resource_groups: Optional[List[CreateK8sResourceGroup]] = None,
    ):

        super().__init__()
        try:
            self.args: ElasticAppArgs = ElasticAppArgs(
                name=name,
                version=version,
                enabled=enabled,
                docker_resource_groups=docker_resource_groups,
                k8s_resource_groups=k8s_resource_groups,
                create_k8s_resource_groups=create_k8s_resource_groups,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        if self.args.docker_resource_groups is not None:
            self.docker_resource_groups = OrderedDict()
            for docker_rg in self.args.docker_resource_groups:
                if docker_rg is not None and isinstance(docker_rg, DockerResourceGroup):
                    self.docker_resource_groups[docker_rg.name] = docker_rg
                else:
                    print_error(
                        "+------+ DockerResourceGroup invalid: {}".format(docker_rg)
                    )

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        # first check if the k8s_resource_groups are provided
        if self.args.k8s_resource_groups is not None:
            self.k8s_resource_groups = OrderedDict()
            for k8s_rg in self.args.k8s_resource_groups:
                if k8s_rg is not None and isinstance(k8s_rg, K8sResourceGroup):
                    self.k8s_resource_groups[k8s_rg.name] = k8s_rg
                else:
                    print_error(
                        "+------+ K8sResourceGroup invalid: {}".format(k8s_rg.name)
                    )
        # second check if the create_k8s_resource_groups are provided
        elif self.args.create_k8s_resource_groups is not None:
            self.k8s_resource_groups = OrderedDict()
            for create_k8s_rg in self.args.create_k8s_resource_groups:
                if create_k8s_rg is not None and isinstance(
                    create_k8s_rg, CreateK8sResourceGroup
                ):
                    _k8s_rg = create_k8s_rg.create()
                    if _k8s_rg is not None:
                        self.k8s_resource_groups[create_k8s_rg.name] = _k8s_rg
                else:
                    print_error(
                        "+------+ K8sResourceGroup invalid: {}".format(
                            create_k8s_rg.name
                        )
                    )



from typing import Optional

from kubernetes.client.models.v1_volume import V1Volume
from pydantic import Field

from phidata.infra.k8s.resource.base import K8sObject
from phidata.infra.k8s.resource.core.v1.volume_source import (
    AwsElasticBlockStoreVolumeSource,
    ConfigMapVolumeSource,
    EmptyDirVolumeSource,
    GcePersistentDiskVolumeSource,
    GitRepoVolumeSource,
    PersistentVolumeClaimVolumeSource,
    SecretVolumeSource,
    HostPathVolumeSource,
)


class Volume(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#volume-v1-core
    """

    resource_type: str = "Volume"

    # Volume's name. Must be a DNS_LABEL and unique within the pod.
    # More info: https://kubernetes.io/docs/concepts/overview/working-with-objects/names/#names
    name: str

    ## Volume Sources
    aws_elastic_block_store: Optional[AwsElasticBlockStoreVolumeSource] = Field(
        None, alias="awsElasticBlockStore"
    )
    # ConfigMap represents a configMap that should populate this volume
    config_map: Optional[ConfigMapVolumeSource] = Field(None, alias="configMap")
    # EmptyDir represents a temporary directory that shares a pod's lifetime.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#emptydir
    empty_dir: Optional[EmptyDirVolumeSource] = Field(None, alias="emptyDir")
    # GCEPersistentDisk represents a GCE Disk resource that is attached to a
    # kubelet's host machine and then exposed to the pod. Provisioned by an admin.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#gcepersistentdisk
    gce_persistent_disk: Optional[GcePersistentDiskVolumeSource] = Field(
        None, alias="gcePersistentDisk"
    )
    # GitRepo represents a git repository at a particular revision.
    # DEPRECATED: GitRepo is deprecated.
    # To provision a container with a git repo, mount an EmptyDir into an InitContainer
    # that clones the repo using git, then mount the EmptyDir into the Pod's container.
    git_repo: Optional[GitRepoVolumeSource] = Field(None, alias="gitRepo")
    # HostPath represents a pre-existing file or directory on the host machine that is
    # directly exposed to the container. This is generally used for system agents or other privileged things
    # that are allowed to see the host machine. Most containers will NOT need this.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#hostpath
    host_path: Optional[HostPathVolumeSource] = Field(None, alias="hostPath")
    # PersistentVolumeClaimVolumeSource represents a reference to a PersistentVolumeClaim in the same namespace.
    # More info: https://kubernetes.io/docs/concepts/storage/persistent-volumes#persistentvolumeclaims
    persistent_volume_claim: Optional[PersistentVolumeClaimVolumeSource] = Field(
        None, alias="persistentVolumeClaim"
    )
    # Secret represents a secret that should populate this volume.
    # More info: https://kubernetes.io/docs/concepts/storage/volumes#secret
    secret: Optional[SecretVolumeSource] = None

    def get_k8s_object(self) -> V1Volume:
        # Return a V1Volume object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_volume.py
        _v1_volume = V1Volume(
            name=self.name,
            aws_elastic_block_store=self.aws_elastic_block_store.get_k8s_object()
            if self.aws_elastic_block_store
            else None,
            # azure_disk=None,
            # azure_file=None,
            # cephfs=None,
            # cinder=None,
            config_map=self.config_map.get_k8s_object() if self.config_map else None,
            # csi=None,
            # downward_api=None,
            empty_dir=self.empty_dir.get_k8s_object() if self.empty_dir else None,
            # ephemeral=None,
            # fc=None,
            # flex_volume=None,
            # flocker=None,
            gce_persistent_disk=self.gce_persistent_disk.get_k8s_object()
            if self.gce_persistent_disk
            else None,
            git_repo=self.git_repo.get_k8s_object() if self.git_repo else None,
            # glusterfs=None,
            host_path=self.host_path.get_k8s_object() if self.host_path else None,
            # iscsi=None,
            # nfs=None,
            persistent_volume_claim=self.persistent_volume_claim.get_k8s_object()
            if self.persistent_volume_claim
            else None,
            # photon_persistent_disk=None,
            # portworx_volume=None,
            # projected=None,
            # quobyte=None,
            # rbd=None,
            # scale_io=None,
            secret=self.secret.get_k8s_object() if self.secret else None,
            # storageos=None,
            # vsphere_volume=None,
        )
        return _v1_volume



from pathlib import Path
from typing import Optional, Any, Literal, cast

from phidata.base import PhidataBase, PhidataBaseArgs
from phidata.constants import STORAGE_DIR_ENV_VAR, PHIDATA_RUNTIME_ENV_VAR
from phidata.utils.log import logger


class DataAssetArgs(PhidataBaseArgs):
    # The base_dir_path is the root dir for the data_asset.
    # Normally corresponds to $WORKSPACE_ROOT/storage
    # This path depends on the environment (local vs container) and is
    # used by the data_asset to build the absolute path in different environments
    base_dir_path: Optional[Path] = None
    phidata_runtime: Optional[Literal["local", "airflow"]] = None


class DataAsset(PhidataBase):
    """Base Class for all DataAssets"""

    def __init__(self) -> None:
        super().__init__()
        self.args: Optional[DataAssetArgs] = None

    @property
    def base_dir_path(self) -> Optional[Path]:
        """
        Returns the base dir for the data_asset.
        This base dir depends on the environment (local vs container) and is
        used by the data_asset to build the absolute path.
        """
        # data_asset not yet initialized
        if self.args is None:
            return None

        if self.args.base_dir_path:
            # use cached value if available
            return self.args.base_dir_path

        import os

        storage_dir_env = os.getenv(STORAGE_DIR_ENV_VAR)
        # logger.debug(f"{STORAGE_DIR_ENV_VAR}: {storage_dir_env}")
        if storage_dir_env is not None:
            self.args.base_dir_path = Path(storage_dir_env)
        return self.args.base_dir_path

    @base_dir_path.setter
    def base_dir_path(self, base_dir_path: Optional[Path]) -> None:
        if self.args is not None and base_dir_path is not None:
            self.args.base_dir_path = base_dir_path

    @property
    def phidata_runtime(self) -> Literal["local", "airflow"]:
        # data_asset not yet initialized
        if self.args is None:
            return "local"

        if self.args.phidata_runtime:
            # use cached value if available
            return self.args.phidata_runtime

        import os

        phidata_runtime_env = os.getenv(PHIDATA_RUNTIME_ENV_VAR)
        # logger.debug(f"{PHIDATA_RUNTIME_ENV_VAR}: {phidata_runtime_env}")
        if phidata_runtime_env is not None and phidata_runtime_env in (
            "local",
            "airflow",
        ):
            phidata_runtime_env = cast(Literal["local", "airflow"], phidata_runtime_env)
            self.args.phidata_runtime = phidata_runtime_env
        else:
            self.args.phidata_runtime = "local"
        return self.args.phidata_runtime

    @phidata_runtime.setter
    def phidata_runtime(self, phidata_runtime: Literal["local", "airflow"]) -> None:
        if (
            self.args is not None
            and phidata_runtime is not None
            and phidata_runtime in ("local", "airflow")
        ):
            self.args.phidata_runtime = phidata_runtime

    ######################################################
    ## Build and validate data asset
    ######################################################

    def build(self) -> bool:
        logger.debug(f"@build not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Write DataAsset
    ######################################################

    def write_pandas_df(self, df: Any = None) -> bool:
        logger.debug(f"@write_pandas_df not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Read DataAsset
    ######################################################

    def read_pandas_df(self) -> Optional[Any]:
        logger.debug(f"@read_pandas_df not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Delete DataAsset
    ######################################################

    def delete(self) -> bool:
        logger.debug(f"@delete not defined for {self.__class__.__name__}")
        return False

    ######################################################
    ## Print
    ######################################################

    def __str__(self) -> str:
        if self.args is not None:
            return self.args.json(indent=2, exclude_none=True, exclude={"enabled"})
        else:
            return self.__class__.__name__



from typing import Dict, List, Optional

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.enums.pv import PVAccessMode
from phidata.infra.k8s.resource.core.v1.persistent_volume_claim import (
    PersistentVolumeClaim,
    PersistentVolumeClaimSpec,
)
from phidata.infra.k8s.resource.core.v1.resource_requirements import (
    ResourceRequirements,
)
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.cli_console import print_info
from phidata.utils.log import logger


class CreatePVC(BaseModel):
    pvc_name: str
    app_name: str
    namespace: Optional[str] = None
    request_storage: str
    storage_class_name: str
    access_modes: List[PVAccessMode] = [PVAccessMode.READ_WRITE_ONCE]
    labels: Optional[Dict[str, str]] = None


def create_pvc_resource(data: CreatePVC) -> Optional[PersistentVolumeClaim]:
    """Creates a PersistentVolumeClaim resource."""

    if data is None:
        return None

    pvc_name = data.pvc_name
    logger.debug(f"Init PersistentVolumeClaim resource: {pvc_name}")

    pvc_labels = create_component_labels_dict(
        component_name=pvc_name,
        app_name=data.app_name,
        labels=data.labels,
    )

    pvc = PersistentVolumeClaim(
        api_version=ApiVersion.CORE_V1,
        kind=Kind.PERSISTENTVOLUMECLAIM,
        metadata=ObjectMeta(
            name=pvc_name,
            namespace=data.namespace,
            labels=pvc_labels,
        ),
        spec=PersistentVolumeClaimSpec(
            access_modes=data.access_modes,
            resources=ResourceRequirements(requests={"storage": data.request_storage}),
            storage_class_name=data.storage_class_name,
        ),
    )

    logger.info(
        f"PersistentVolumeClaim {pvc_name}:\n{pvc.json(exclude_defaults=True, indent=2)}"
    )
    return pvc



PHIDATA_RUNTIME_ENV_VAR: str = "PHIDATA_RUNTIME"

SCRIPTS_DIR_ENV_VAR: str = "PHI_SCRIPTS_DIR"
STORAGE_DIR_ENV_VAR: str = "PHI_STORAGE_DIR"
META_DIR_ENV_VAR: str = "PHI_META_DIR"
PRODUCTS_DIR_ENV_VAR: str = "PHI_PRODUCTS_DIR"
NOTEBOOKS_DIR_ENV_VAR: str = "PHI_NOTEBOOKS_DIR"
WORKSPACE_CONFIG_DIR_ENV_VAR: str = "PHI_WORKSPACE_CONFIG_DIR"

AWS_REGION_ENV_VAR: str = "AWS_REGION"
AWS_PROFILE_ENV_VAR: str = "AWS_PROFILE"
AWS_CONFIG_FILE_ENV_VAR: str = "AWS_CONFIG_FILE"
AWS_SHARED_CREDENTIALS_FILE_ENV_VAR: str = "AWS_SHARED_CREDENTIALS_FILE"



from pathlib import Path
from typing import Optional, Dict
from typing_extensions import Literal

from phidata.base import PhidataBase
from phidata.infra.base.args import InfraArgs
from phidata.utils.log import logger


class InfraConfig(PhidataBase):
    """Base Class for all phidata infra configs"""

    def __init__(self) -> None:
        super().__init__()
        self.args: Optional[InfraArgs] = None

    @property
    def env(self) -> Optional[Literal["dev", "stg", "prd"]]:
        return self.args.env if self.args else None

    @property
    def workspace_root_path(self) -> Optional[Path]:
        return self.args.workspace_root_path if self.args else None

    @workspace_root_path.setter
    def workspace_root_path(self, workspace_root_path: Path) -> None:
        if self.args is not None and workspace_root_path is not None:
            self.args.workspace_root_path = workspace_root_path

    @property
    def workspace_config_file_path(self) -> Optional[Path]:
        return self.args.workspace_config_file_path if self.args else None

    @workspace_config_file_path.setter
    def workspace_config_file_path(self, workspace_config_file_path: Path) -> None:
        if self.args is not None and workspace_config_file_path is not None:
            self.args.workspace_config_file_path = workspace_config_file_path

    @property
    def scripts_dir(self) -> Optional[str]:
        return self.args.scripts_dir if self.args else None

    @scripts_dir.setter
    def scripts_dir(self, scripts_dir: str) -> None:
        if self.args is not None and scripts_dir is not None:
            self.args.scripts_dir = scripts_dir

    @property
    def storage_dir(self) -> Optional[str]:
        return self.args.storage_dir if self.args else None

    @storage_dir.setter
    def storage_dir(self, storage_dir: str) -> None:
        if self.args is not None and storage_dir is not None:
            self.args.storage_dir = storage_dir

    @property
    def meta_dir(self) -> Optional[str]:
        return self.args.meta_dir if self.args else None

    @meta_dir.setter
    def meta_dir(self, meta_dir: str) -> None:
        if self.args is not None and meta_dir is not None:
            self.args.meta_dir = meta_dir

    @property
    def products_dir(self) -> Optional[str]:
        return self.args.products_dir if self.args else None

    @products_dir.setter
    def products_dir(self, products_dir: str) -> None:
        if self.args is not None and products_dir is not None:
            self.args.products_dir = products_dir

    @property
    def notebooks_dir(self) -> Optional[str]:
        return self.args.notebooks_dir if self.args else None

    @notebooks_dir.setter
    def notebooks_dir(self, notebooks_dir: str) -> None:
        if self.args is not None and notebooks_dir is not None:
            self.args.notebooks_dir = notebooks_dir

    @property
    def workspace_config_dir(self) -> Optional[str]:
        return self.args.workspace_config_dir if self.args else None

    @workspace_config_dir.setter
    def workspace_config_dir(self, workspace_config_dir: str) -> None:
        if self.args is not None and workspace_config_dir is not None:
            self.args.workspace_config_dir = workspace_config_dir

    @property
    def local_env(self) -> Optional[Dict[str, str]]:
        return self.args.local_env if self.args else None

    @local_env.setter
    def local_env(self, local_env: Dict[str, str]) -> None:
        if self.args is not None and local_env is not None:
            self.args.local_env = local_env

    @property
    def docker_env(self) -> Optional[Dict[str, str]]:
        return self.args.docker_env if self.args else None

    @docker_env.setter
    def docker_env(self, docker_env: Dict[str, str]) -> None:
        if self.args is not None and docker_env is not None:
            self.args.docker_env = docker_env

    @property
    def k8s_env(self) -> Optional[Dict[str, str]]:
        return self.args.k8s_env if self.args else None

    @k8s_env.setter
    def k8s_env(self, k8s_env: Dict[str, str]) -> None:
        if self.args is not None and k8s_env is not None:
            self.args.k8s_env = k8s_env

    @property
    def aws_region(self) -> Optional[str]:
        return self.args.aws_region if self.args else None

    @aws_region.setter
    def aws_region(self, aws_region: str) -> None:
        if self.args is not None and aws_region is not None:
            self.args.aws_region = aws_region

    @property
    def aws_profile(self) -> Optional[str]:
        return self.args.aws_profile if self.args else None

    @aws_profile.setter
    def aws_profile(self, aws_profile: str) -> None:
        if self.args is not None and aws_profile is not None:
            self.args.aws_profile = aws_profile

    @property
    def aws_config_file(self) -> Optional[str]:
        return self.args.aws_config_file if self.args else None

    @aws_config_file.setter
    def aws_config_file(self, aws_config_file: str) -> None:
        if self.args is not None and aws_config_file is not None:
            self.args.aws_config_file = aws_config_file

    @property
    def aws_shared_credentials_file(self) -> Optional[str]:
        return self.args.aws_shared_credentials_file if self.args else None

    @aws_shared_credentials_file.setter
    def aws_shared_credentials_file(self, aws_shared_credentials_file: str) -> None:
        if self.args is not None and aws_shared_credentials_file is not None:
            self.args.aws_shared_credentials_file = aws_shared_credentials_file

    def is_valid(self) -> bool:
        """
        This function is implemented by each subclass
        and called before the config is used
        """
        logger.debug(f"@is_valid not defined for {self.__class__.__name__}")
        return True



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, List, Union
from typing_extensions import Literal

from phidata.app.db import DbApp, DbAppArgs
from phidata.infra.aws.resource.ec2.volume import EbsVolume
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment, RestartPolicy
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer, ImagePullPolicy
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.service import CreateService, ServiceType
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    HostPathVolumeSource,
    AwsElasticBlockStoreVolumeSource,
    VolumeType,
)
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.cli_console import print_error, print_info, print_warning
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_service_name,
    get_default_volume_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class RedisVolumeType(ExtendedEnum):
    EMPTY_DIR = "EMPTY_DIR"
    HOST_PATH = "HOST_PATH"
    PERSISTENT_VOLUME = "PERSISTENT_VOLUME"
    AWS_EBS = "AWS_EBS"


class RedisArgs(DbAppArgs):
    name: str = "redis"
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "redis"
    image_tag: str = "6.2.6"
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # Configure redis
    # Provide REDIS_PASSWORD as redis_password or REDIS_PASSWORD in secrets_file
    redis_password: Optional[str] = None
    # Provide REDIS_SCHEMA as redis_schema or REDIS_SCHEMA in secrets_file
    redis_schema: Optional[str] = None
    redis_driver: str = "redis"
    logging_level: Literal[
        "debug",
        "info",
        "warning",
        "error",
        "critical",
    ] = "debug"

    # Configure the volume
    create_volume: bool = True
    volume_type: RedisVolumeType = RedisVolumeType.EMPTY_DIR
    volume_name: Optional[str] = None
    # Container path to mount the postgres volume
    # should be the parent directory of redisdata
    volume_container_path: str = "/data"
    # Host path to mount the postgres volume
    # If volume_type = RedisVolumeType.HOST_PATH
    volume_host_path: Optional[Path] = None
    # EbsVolume if volume_type = PostgresVolumeType.AWS_EBS
    ebs_volume: Optional[EbsVolume] = None
    # EbsVolume region is used to determine the ebs_volume_id
    # and add topology region selectors
    ebs_volume_region: Optional[str] = None
    # Provide Ebs Volume-id manually
    ebs_volume_id: Optional[str] = None
    # Add topology az selectors
    ebs_volume_az: Optional[str] = None
    # Add NodeSelectors to Pods, so they are scheduled in the same
    # region and zone as the ebs_volume
    schedule_pods_in_ebs_topology: bool = True

    # Configure the container
    container_name: Optional[str] = None
    image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT
    container_port: int = 6379
    # Only used by the K8sContainer
    container_port_name: str = "redis"
    # Only used by the DockerContainer
    container_host_port: int = 6379
    container_detach: bool = True
    container_auto_remove: bool = True
    container_remove: bool = True

    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Configure the ConfigMap used for env variables that are not Secret
    config_map_name: Optional[str] = None
    # Configure the Secret used for env variables that are Secret
    secret_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # Configure the deployment
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None
    replicas: int = 1
    pod_node_selector: Optional[Dict[str, str]] = None
    restart_policy: RestartPolicy = RestartPolicy.ALWAYS
    termination_grace_period_seconds: Optional[int] = None

    # Configure the service
    service_name: Optional[str] = None
    service_type: Optional[ServiceType] = None
    # The port that will be exposed by the service.
    service_port: int = 6379
    # The node_port that will be exposed by the service if service_type = ServiceType.NODE_PORT
    node_port: Optional[int] = None
    # The target_port is the port to access on the pods targeted by the service.
    # It can be the port number or port name on the pod.
    target_port: Optional[Union[str, int]] = None


class Redis(DbApp):
    def __init__(
        self,
        name: str = "redis",
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "redis",
        image_tag: str = "6.2.6",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Configure redis,
        # Provide REDIS_PASSWORD as redis_password or REDIS_PASSWORD in secrets_file,
        redis_password: Optional[str] = None,
        # Provide REDIS_SCHEMA as redis_schema or REDIS_SCHEMA in secrets_file,
        redis_schema: Optional[str] = None,
        redis_driver: str = "redis",
        logging_level: Literal[
            "debug",
            "info",
            "warning",
            "error",
            "critical",
        ] = "debug",
        # Configure the volume,
        create_volume: bool = True,
        volume_type: RedisVolumeType = RedisVolumeType.EMPTY_DIR,
        volume_name: Optional[str] = None,
        # Container path to mount the postgres volume,
        # should be the parent directory of redisdata,
        volume_container_path: str = "/data",
        # Host path to mount the postgres volume,
        # If volume_type = RedisVolumeType.HOST_PATH,
        volume_host_path: Optional[Path] = None,
        # EbsVolume if volume_type = PostgresVolumeType.AWS_EBS,
        ebs_volume: Optional[EbsVolume] = None,
        # EbsVolume region is used to determine the ebs_volume_id,
        # and add topology region selectors,
        ebs_volume_region: Optional[str] = None,
        # Provide Ebs Volume-id manually,
        ebs_volume_id: Optional[str] = None,
        # Add topology az selectors,
        ebs_volume_az: Optional[str] = None,
        # Add NodeSelectors to Pods, so they are scheduled in the same,
        # region and zone as the ebs_volume,
        schedule_pods_in_ebs_topology: bool = True,
        # Configure the container,
        container_name: Optional[str] = None,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        container_port: int = 6379,
        # Only used by the K8sContainer,
        container_port_name: str = "redis",
        # Only used by the DockerContainer,
        container_host_port: int = 6379,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the deployment,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Configure the service,
        service_name: Optional[str] = None,
        service_type: Optional[ServiceType] = None,
        # The port that will be exposed by the service.,
        service_port: int = 6379,
        # The node_port that will be exposed by the service if service_type = ServiceType.NODE_PORT,
        node_port: Optional[int] = None,
        # The target_port is the port to access on the pods targeted by the service.,
        # It can be the port number or port name on the pod.,
        target_port: Optional[Union[str, int]] = None,
        # Additional args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        # If True, log extra debug messages
        use_verbose_logs: bool = False,
    ):
        super().__init__()
        try:
            self.args: RedisArgs = RedisArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                redis_password=redis_password,
                redis_schema=redis_schema,
                redis_driver=redis_driver,
                logging_level=logging_level,
                create_volume=create_volume,
                volume_type=volume_type,
                volume_name=volume_name,
                volume_container_path=volume_container_path,
                volume_host_path=volume_host_path,
                ebs_volume=ebs_volume,
                ebs_volume_region=ebs_volume_region,
                ebs_volume_id=ebs_volume_id,
                ebs_volume_az=ebs_volume_az,
                schedule_pods_in_ebs_topology=schedule_pods_in_ebs_topology,
                container_name=container_name,
                image_pull_policy=image_pull_policy,
                container_port=container_port,
                container_port_name=container_port_name,
                container_host_port=container_host_port,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                deploy_name=deploy_name,
                pod_name=pod_name,
                replicas=replicas,
                pod_node_selector=pod_node_selector,
                restart_policy=restart_policy,
                termination_grace_period_seconds=termination_grace_period_seconds,
                service_name=service_name,
                service_type=service_type,
                service_port=service_port,
                node_port=node_port,
                target_port=target_port,
                use_cache=use_cache,
                use_verbose_logs=use_verbose_logs,
            )
        except Exception:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_service_name(self) -> str:
        return self.args.service_name or get_default_service_name(self.args.name)

    def get_service_port(self) -> int:
        return self.args.service_port

    def get_env_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        env_file_path = self.args.env_file
        if (
            env_file_path is not None
            and env_file_path.exists()
            and env_file_path.is_file()
        ):
            # logger.debug(f"Reading {env_file_path}")
            env_data_from_file = yaml.safe_load(env_file_path.read_text())
            if env_data_from_file is not None and isinstance(env_data_from_file, dict):
                return env_data_from_file
            else:
                print_error(f"Invalid env_file: {env_file_path}")
        return None

    def get_secret_data_from_file(self) -> Optional[Dict[str, str]]:
        import yaml

        secrets_file_path = self.args.secrets_file
        if (
            secrets_file_path is not None
            and secrets_file_path.exists()
            and secrets_file_path.is_file()
        ):
            # logger.debug(f"Reading {secrets_file_path}")
            secret_data_from_file = yaml.safe_load(secrets_file_path.read_text())
            if secret_data_from_file is not None and isinstance(
                secret_data_from_file, dict
            ):
                return secret_data_from_file
            else:
                print_error(f"Invalid secrets_file: {secrets_file_path}")
        return None

    def get_db_password(self) -> Optional[str]:
        redis_password_var: Optional[str] = self.args.db_password if self.args else None
        if redis_password_var is None and self.args.secrets_file is not None:
            # read from secrets_file
            logger.debug(f"Reading REDIS_PASSWORD from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                redis_password_var = secret_data_from_file.get(
                    "REDIS_PASSWORD", redis_password_var
                )
        return redis_password_var

    def get_db_schema(self) -> Optional[str]:
        redis_schema_var: Optional[str] = self.args.db_schema if self.args else None
        if redis_schema_var is None and self.args.secrets_file is not None:
            # read from env_file
            logger.debug(f"Reading REDIS_SCHEMA from secrets_file")
            secret_data_from_file = self.get_secret_data_from_file()
            if secret_data_from_file is not None:
                redis_schema_var = secret_data_from_file.get(
                    "REDIS_SCHEMA", redis_schema_var
                )
        return redis_schema_var

    def get_db_driver(self) -> Optional[str]:
        return self.args.redis_driver if self.args else "redis"

    def get_db_host_local(self) -> Optional[str]:
        return "localhost"

    def get_db_port_local(self) -> Optional[int]:
        return self.args.container_host_port if self.args else None

    def get_db_host_docker(self) -> Optional[str]:
        return self.get_container_name()

    def get_db_port_docker(self) -> Optional[int]:
        return self.args.container_port if self.args else None

    def get_db_host_k8s(self) -> Optional[str]:
        return self.get_service_name()

    def get_db_port_k8s(self) -> Optional[int]:
        return self.get_service_port()

    def get_db_connection_url_local(self) -> Optional[str]:
        password = self.get_db_password()
        password_str = f"{password}@" if password else ""
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_local()
        port = self.get_db_port_local()
        return f"{driver}://{password_str}@{host}:{port}/{schema}"

    def get_db_connection_url_docker(self) -> Optional[str]:
        password = self.get_db_password()
        password_str = f"{password}@" if password else ""
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_docker()
        port = self.get_db_port_docker()
        return f"{driver}://{password_str}@{host}:{port}/{schema}"

    def get_db_connection_url_k8s(self) -> Optional[str]:
        password = self.get_db_password()
        password_str = f"{password}@" if password else ""
        schema = self.get_db_schema()
        driver = self.get_db_driver()
        host = self.get_db_host_k8s()
        port = self.get_db_port_k8s()
        return f"{driver}://{password_str}@{host}:{port}/{schema}"

    ######################################################
    ## Docker Resources
    ######################################################

    def get_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} DockerResourceGroup")

        # Container Environment
        container_env: Dict[str, str] = {}

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env.update(secret_data_from_file)

        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Container Volumes
        container_volumes = {}
        if self.args.create_volume:
            if self.args.volume_type == RedisVolumeType.EMPTY_DIR:
                volume_name = self.args.volume_name or get_default_volume_name(app_name)
                container_volumes[volume_name] = {
                    "bind": self.args.volume_container_path,
                    "mode": "rw",
                }
            elif self.args.volume_type == RedisVolumeType.HOST_PATH:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    container_volumes[volume_host_path_str] = {
                        "bind": self.args.volume_container_path,
                        "mode": "rw",
                    }
                else:
                    print_error("Redis: volume_host_path not provided")
                    return None
            else:
                print_error(f"{self.args.volume_type.value} not supported")
                return None

        # Container Ports
        container_ports: Dict[str, int] = {
            str(self.args.container_port): self.args.container_host_port,
        }

        # Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=get_image_str(self.args.image_name, self.args.image_tag),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            remove=self.args.container_remove,
            stdin_open=True,
            tty=True,
            environment=container_env,
            network=docker_build_context.network,
            ports=container_ports,
            volumes=container_volumes,
            use_cache=self.args.use_cache,
        )

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        # Container Environment
        container_env: Dict[str, str] = {}

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)
        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)
        # Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.config_map_name or get_default_configmap_name(app_name),
            app_name=app_name,
            data=container_env,
        )

        # Create a Secret to set the container env variables which are Secret
        container_env_secret: Optional[CreateSecret] = None
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env_secret = CreateSecret(
                secret_name=self.args.secret_name or get_default_secret_name(app_name),
                app_name=app_name,
                string_data=secret_data_from_file,
            )

        # Container Volumes
        container_volumes = []
        # Add NodeSelectors to Pods in case we create az sensitive volumes
        pod_node_selector: Optional[Dict[str, str]] = self.args.pod_node_selector
        if self.args.create_volume:
            volume_name = self.args.volume_name or get_default_volume_name(app_name)
            if self.args.volume_type == RedisVolumeType.EMPTY_DIR:
                redis_volume = CreateVolume(
                    volume_name=volume_name,
                    app_name=app_name,
                    mount_path=self.args.volume_container_path,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                container_volumes.append(redis_volume)
            elif self.args.volume_type == RedisVolumeType.HOST_PATH:
                if self.args.volume_host_path is not None:
                    volume_host_path_str = str(self.args.volume_host_path)
                    redis_volume = CreateVolume(
                        volume_name=volume_name,
                        app_name=app_name,
                        mount_path=self.args.volume_container_path,
                        volume_type=VolumeType.HOST_PATH,
                        host_path=HostPathVolumeSource(
                            path=volume_host_path_str,
                        ),
                    )
                    container_volumes.append(redis_volume)
                else:
                    print_error("Redis: volume_host_path not provided")
                    return None
            elif self.args.volume_type == RedisVolumeType.AWS_EBS:
                if (
                    self.args.ebs_volume_id is not None
                    or self.args.ebs_volume is not None
                ):
                    ebs_volume_id = self.args.ebs_volume_id
                    if ebs_volume_id is None and self.args.ebs_volume is not None:
                        ebs_volume_id = self.args.ebs_volume.get_volume_id(
                            aws_region=self.args.ebs_volume_region
                        )
                    logger.debug(f"ebs_volume_id: {ebs_volume_id}")

                    redis_volume = CreateVolume(
                        volume_name=volume_name,
                        app_name=app_name,
                        mount_path=self.args.volume_container_path,
                        volume_type=VolumeType.AWS_EBS,
                        aws_ebs=AwsElasticBlockStoreVolumeSource(
                            volume_id=ebs_volume_id,
                        ),
                    )
                    container_volumes.append(redis_volume)
                    if self.args.schedule_pods_in_ebs_topology:
                        if pod_node_selector is None:
                            pod_node_selector = {}

                        # Add NodeSelectors to Pods, so they are scheduled in the same
                        # region and zone as the ebs_volume
                        # https://kubernetes.io/docs/reference/labels-annotations-taints/#topologykubernetesiozone
                        ebs_region = self.args.ebs_volume_region
                        if ebs_region is not None:
                            pod_node_selector[
                                "topology.kubernetes.io/region"
                            ] = ebs_region

                        ebs_az = self.args.ebs_volume_az
                        if ebs_az is None and self.args.ebs_volume is not None:
                            ebs_az = self.args.ebs_volume.availability_zone
                        if ebs_az is not None:
                            pod_node_selector["topology.kubernetes.io/zone"] = ebs_az
                else:
                    print_error("Redis: ebs_volume not provided")
                    return None
            else:
                print_error(f"{self.args.volume_type.value} not supported")
                return None

        # Create the ports to open
        container_port = CreatePort(
            name=self.args.container_port_name,
            container_port=self.args.container_port,
            service_port=self.args.service_port,
            target_port=self.args.target_port or self.args.container_port_name,
        )

        # If ServiceType == NODE_PORT then validate self.args.node_port is available
        if self.args.service_type == ServiceType.NODE_PORT:
            if (
                self.args.node_port is None
                or self.args.node_port < 30000
                or self.args.node_port > 32767
            ):
                print_error(f"NodePort: {self.args.node_port} invalid")
                print_error(f"Skipping this service")
                return None
            else:
                container_port.node_port = self.args.node_port
        # If ServiceType == LOAD_BALANCER then validate self.args.node_port only IF available
        elif self.args.service_type == ServiceType.LOAD_BALANCER:
            if self.args.node_port is not None:
                if self.args.node_port < 30000 or self.args.node_port > 32767:
                    print_error(f"NodePort: {self.args.node_port} invalid")
                    print_error(f"Skipping this service")
                    return None
                else:
                    container_port.node_port = self.args.node_port
        # else validate self.args.node_port is NOT available
        elif self.args.node_port is not None:
            print_warning(
                f"NodePort: {self.args.node_port} provided without specifying ServiceType as NODE_PORT or LOAD_BALANCER"
            )
            print_warning("NodePort value will be ignored")
            self.args.node_port = None

        # Create the container
        k8s_container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=[self.args.command]
            if isinstance(self.args.command, str)
            else self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            image_pull_policy=self.args.image_pull_policy,
            envs_from_configmap=[container_env_cm.cm_name],
            envs_from_secret=[container_env_secret.secret_name]
            if container_env_secret
            else None,
            ports=[container_port],
            volumes=container_volumes,
            labels=k8s_build_context.labels,
        )

        # Create the deployment
        k8s_deployment = CreateDeployment(
            deploy_name=self.args.deploy_name or get_default_deploy_name(app_name),
            pod_name=self.args.pod_name or get_default_pod_name(app_name),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            replicas=self.args.replicas,
            containers=[k8s_container],
            pod_node_selector=pod_node_selector,
            restart_policy=self.args.restart_policy,
            termination_grace_period_seconds=self.args.termination_grace_period_seconds,
            volumes=container_volumes,
            labels=k8s_build_context.labels,
        )

        # Create the service
        k8s_service = CreateService(
            service_name=self.get_service_name(),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            service_type=self.args.service_type,
            deployment=k8s_deployment,
            ports=[container_port],
            labels=k8s_build_context.labels,
        )

        # Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            config_maps=[container_env_cm],
            secrets=[container_env_secret] if container_env_secret else None,
            services=[k8s_service],
            deployments=[k8s_deployment],
        )

        return k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg = self.get_k8s_rg(k8s_build_context)
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



from kubernetes.client.models.v1_pod_template_spec import V1PodTemplateSpec

from phidata.infra.k8s.resource.base import K8sObject
from phidata.infra.k8s.resource.core.v1.pod_spec import PodSpec
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta


class PodTemplateSpec(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#podtemplatespec-v1-core
    """

    resource_type: str = "PodTemplateSpec"

    metadata: ObjectMeta
    spec: PodSpec

    def get_k8s_object(self) -> V1PodTemplateSpec:

        # Set and return a V1PodTemplateSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_pod_template_spec.py
        _v1_pod_template_spec = V1PodTemplateSpec(
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object(),
        )
        return _v1_pod_template_spec



from typing import Optional, Any, Union, List, Dict, Tuple
from typing_extensions import Literal

from sqlalchemy.engine import Engine, Connection
from sqlalchemy.exc import ResourceClosedError

from phidata.asset import DataAsset, DataAssetArgs
from phidata.utils.enums import ExtendedEnum
from phidata.utils.log import logger


class SqlType(ExtendedEnum):
    POSTGRES = "POSTGRES"


class SqlTableArgs(DataAssetArgs):
    # Table Name
    name: str
    # Type of SQL table
    sql_type: SqlType
    # Table schema
    db_schema: Optional[str] = None
    # airflow connection_id can be used for running workflows remotely
    db_conn_id: Optional[str] = None
    # a db_conn_url can be used to create the sqlalchemy.engine.Engine object
    db_conn_url: Optional[str] = None
    # sqlalchemy.engine.(Engine or Connection)
    # Using SQLAlchemy makes it possible to use any DB supported by that library.
    # NOTE: db_engine is required but can be derived using
    # db_conn_url when running locally and
    # db_conn_id when running remotely
    db_engine: Optional[Union[Engine, Connection]] = None


class SqlTable(DataAsset):
    """Base Class for Sql tables"""

    def __init__(
        self,
        name: Optional[str] = None,
        sql_type: Optional[SqlType] = None,
        db_schema: Optional[str] = None,
        db_conn_id: Optional[str] = None,
        db_conn_url: Optional[str] = None,
        db_engine: Optional[Union[Engine, Connection]] = None,
    ) -> None:
        super().__init__()
        self.args: Optional[SqlTableArgs] = None
        if name is not None and sql_type is not None:
            try:
                self.args = SqlTableArgs(
                    name=name,
                    sql_type=sql_type,
                    db_schema=db_schema,
                    db_conn_id=db_conn_id,
                    db_conn_url=db_conn_url,
                    db_engine=db_engine,
                )
            except Exception as e:
                logger.error(f"Args for {self.__class__.__name__} are not valid")
                raise

    @property
    def sql_type(self) -> Optional[SqlType]:
        return self.args.sql_type if self.args else None

    @sql_type.setter
    def sql_type(self, sql_type: SqlType) -> None:
        if self.args and sql_type:
            self.args.sql_type = sql_type

    @property
    def db_conn_id(self) -> Optional[str]:
        return self.args.db_conn_id if self.args else None

    @db_conn_id.setter
    def db_conn_id(self, db_conn_id: str) -> None:
        if self.args and db_conn_id:
            self.args.db_conn_id = db_conn_id

    @property
    def db_schema(self) -> Optional[str]:
        return self.args.db_schema if self.args else None

    @db_schema.setter
    def db_schema(self, db_schema: str) -> None:
        if self.args and db_schema:
            self.args.db_schema = db_schema

    @property
    def db_conn_url(self) -> Optional[str]:
        return self.args.db_conn_url if self.args else None

    @db_conn_url.setter
    def db_conn_url(self, db_conn_url: str) -> None:
        if self.args and db_conn_url:
            self.args.db_conn_url = db_conn_url

    @property
    def db_engine(self) -> Optional[Union[Engine, Connection]]:
        return self.args.db_engine if self.args else None

    @db_engine.setter
    def db_engine(self, db_engine: Union[Engine, Connection]) -> None:
        if self.args and db_engine:
            self.args.db_engine = db_engine

    def create_db_engine_using_conn_url(self) -> None:
        # Create the SQLAlchemy engine using db_conn_url

        if self.db_conn_url is None:
            return

        try:
            from sqlalchemy import create_engine

            logger.info(f"Creating db_engine using db_conn_url: {self.db_conn_url}")
            db_engine = create_engine(self.db_conn_url)
            logger.debug(f"db_engine: {db_engine}")
            if isinstance(db_engine, tuple) and len(db_engine) > 0:
                self.db_engine = db_engine[0]
            else:
                self.db_engine = db_engine
        except Exception as e:
            logger.error(f"Error creating db_engine using {self.db_conn_url}")
            logger.error(e)
            return

    def create_db_engine_using_conn_id(self) -> None:
        # Create the SQLAlchemy engine using db_conn_id

        if self.db_conn_id is None:
            return

        try:
            from airflow.providers.postgres.hooks.postgres import PostgresHook

            logger.info(f"Creating db_engine using db_conn_id: {self.db_conn_id}")
            if self.sql_type == SqlType.POSTGRES:
                pg_hook = PostgresHook(postgres_conn_id=self.db_conn_id)
                self.db_engine = pg_hook.get_sqlalchemy_engine()
        except Exception as e:
            logger.error(f"Error creating db_engine using {self.db_conn_id}")
            logger.error(e)
            return

    def create_db_engine(self) -> None:
        if self.db_engine is not None:
            return

        phidata_runtime: Literal["local", "airflow"] = self.phidata_runtime
        if phidata_runtime == "local":
            self.create_db_engine_using_conn_url()
        elif phidata_runtime == "airflow":
            self.create_db_engine_using_conn_id()

    ######################################################
    ## Write table
    ######################################################

    def write_pandas_df(
        self,
        df: Optional[Any] = None,
        # How to behave if the table already exists.
        # fail: Raise a ValueError.
        # replace: Drop the table before inserting new values.
        # append: Insert new values to the existing table.
        if_exists: Optional[Literal["fail", "replace", "append"]] = None,
        # Write DataFrame index as a column.
        # Uses index_label as the column name in the table.
        index: Optional[bool] = None,
        # Column label for index column(s).
        # If None is given (default) and index is True, then the index names are used.
        # A sequence should be given if the DataFrame uses MultiIndex.
        index_label: Optional[Union[str, List[str]]] = None,
        # Specify the number of rows in each batch to be written at a time.
        # By default, all rows will be written at once.
        chunksize: Optional[int] = None,
    ) -> bool:
        """
        Write DataFrame to table.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return False

        # Check name is available
        if self.name is None:
            logger.error("SqlTable name not available")
            return False

        # Check engine is available
        if self.db_engine is None:
            self.create_db_engine()
            if self.db_engine is None:
                logger.error("DbEngine not available")
                return False

        # write to table
        import pandas as pd

        if df is None or not isinstance(df, pd.DataFrame):
            logger.error("DataFrame invalid")
            return False

        rows_in_df = df.shape[0]
        logger.info(f"Writing {rows_in_df} rows to table: {self.name}")

        # create a dict of args provided
        not_null_args: Dict[str, Any] = {}
        if self.db_schema:
            not_null_args["schema"] = self.db_schema
        if if_exists:
            not_null_args["if_exists"] = if_exists
        if index:
            not_null_args["index"] = index
        if index_label:
            not_null_args["index_label"] = index_label
        if chunksize:
            not_null_args["chunksize"] = chunksize

        try:
            with self.db_engine.connect() as connection:
                df.to_sql(
                    name=self.name,
                    con=connection,
                    **not_null_args,
                )
                logger.info(f"--**-- Done")
            return True
        except Exception:
            logger.error("Could not write table: {}".format(self.name))
            raise

    ######################################################
    ## Read table
    ######################################################

    def read_pandas_df(
        self,
        index_col: Optional[Union[str, List[str]]] = None,
        coerce_float: bool = True,
        parse_dates: Optional[Union[List, Dict]] = None,
        columns: Optional[List[str]] = None,
        chunksize: Optional[int] = None,
    ) -> Optional[Any]:
        """
        Read table into a DataFrame.

        Args:
            index_col : str or list of str, optional, default: None
                Column(s) to set as index(MultiIndex).
            coerce_float : bool, default True
                Attempts to convert values of non-string, non-numeric objects (like
                decimal.Decimal) to floating point. Can result in loss of Precision.
            parse_dates : list or dict, default None
                - List of column names to parse as dates.
                - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
                - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
            columns : list, default None
                List of column names to select from SQL table.
            chunksize : int, default None
                If specified, returns an iterator where `chunksize` is the number of
                rows to include in each chunk.

        Returns:
            DataFrame or Iterator[DataFrame]
            A SQL table is returned as two-dimensional data structure with labeled
            axes.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return False

        # Check db_engine is available
        if self.db_engine is None:
            self.create_db_engine()
            if self.db_engine is None:
                logger.error("DbEngine not available")
                return False

        # read sql table
        import pandas as pd

        logger.info("Reading table: {}".format(self.name))

        # create a dict of args provided
        not_null_args: Dict[str, Any] = {}
        if self.db_schema:
            not_null_args["schema"] = self.db_schema
        if index_col:
            not_null_args["index_col"] = index_col
        if coerce_float:
            not_null_args["coerce_float"] = coerce_float
        if parse_dates:
            not_null_args["parse_dates"] = parse_dates
        if columns:
            not_null_args["columns"] = columns
        if chunksize:
            not_null_args["chunksize"] = chunksize

        try:
            with self.db_engine.connect() as connection:
                result_df = pd.read_sql_table(
                    table_name=self.name,
                    con=connection,
                    **not_null_args,
                )
            return result_df
        except Exception:
            logger.error(f"Could not read table: {self.name}")
            raise

    def run_sql_query(
        self,
        sql_query: str,
        index_col: Optional[Union[str, List[str]]] = None,
        coerce_float: bool = True,
        parse_dates: Optional[Union[List, Dict]] = None,
        columns: Optional[List[str]] = None,
        chunksize: Optional[int] = None,
    ) -> Optional[Any]:
        """
        Run SQL query using pandas.read_sql()

        Args:
            index_col : str or list of str, optional, default: None
                Column(s) to set as index(MultiIndex).
            coerce_float : bool, default True
                Attempts to convert values of non-string, non-numeric objects (like
                decimal.Decimal) to floating point. Can result in loss of Precision.
            parse_dates : list or dict, default None
                - List of column names to parse as dates.
                - Dict of ``{column_name: format string}`` where format string is
                strftime compatible in case of parsing string times or is one of
                (D, s, ns, ms, us) in case of parsing integer timestamps.
                - Dict of ``{column_name: arg dict}``, where the arg dict corresponds
                to the keyword arguments of :func:`pandas.to_datetime`
                Especially useful with databases without native Datetime support,
                such as SQLite.
            columns : list, default None
                List of column names to select from SQL table.
            chunksize : int, default None
                If specified, returns an iterator where `chunksize` is the number of
                rows to include in each chunk.

        Returns:
            DataFrame or Iterator[DataFrame]
            A SQL table is returned as two-dimensional data structure with labeled
            axes.
        """

        # SqlTable not yet initialized
        if self.args is None:
            return None

        # Check db_engine is available
        if self.db_engine is None:
            self.create_db_engine()
            if self.db_engine is None:
                logger.error("DbEngine not available")
                return None

        # run sql query
        import pandas as pd

        logger.info("Running Query:\n{}".format(sql_query))

        # create a dict of args provided
        not_null_args: Dict[str, Any] = {}
        if self.db_schema:
            not_null_args["schema"] = self.db_schema
        if index_col:
            not_null_args["index_col"] = index_col
        if coerce_float:
            not_null_args["coerce_float"] = coerce_float
        if parse_dates:
            not_null_args["parse_dates"] = parse_dates
        if columns:
            not_null_args["columns"] = columns
        if chunksize:
            not_null_args["chunksize"] = chunksize

        try:
            with self.db_engine.connect() as connection:
                result_df = pd.read_sql(
                    sql=sql_query,
                    con=connection,
                    **not_null_args,
                )
            return result_df
        except ResourceClosedError as rce:
            logger.info(
                f"The result object was closed automatically, returning no rows."
            )
        except Exception as e:
            logger.error(f"Sql query failed: {e}")
        return None

    ######################################################
    ## Drop table
    ######################################################

    def delete(self) -> bool:
        try:
            result = self.run_sql_query(f"DROP TABLE {self.name};")
        except Exception as e:
            logger.error(f"Drop table failed: {e}")
        return True



from typing import Optional, List
from phidata.task.task import Task


class TaskRelatives:
    def __init__(
        self,
        upstream_list: Optional[List[Task]] = None,
        downstream_list: Optional[List[Task]] = None,
    ):
        # list of tasks that run before i.e. parent tasks
        self.upstream_list: Optional[List[Task]] = upstream_list
        # list of tasks that run after i.e. child tasks
        self.downstream_list: Optional[List[Task]] = downstream_list

    def add_upstream(self, task: Task):
        if self.upstream_list is None:
            self.upstream_list = []
        self.upstream_list.append(task)

    def add_upstream_list(self, tasks: List[Task]):
        for task in tasks:
            self.add_upstream(task)

    def add_downstream(self, task: Task):
        if self.downstream_list is None:
            self.downstream_list = []
        self.downstream_list.append(task)

    def add_downstream_list(self, tasks: List[Task]):
        for task in tasks:
            self.add_downstream(task)



from typing import Dict, Optional, List

from pydantic import BaseModel

from phidata.infra.k8s.enums.api_version import ApiVersion
from phidata.infra.k8s.enums.kind import Kind
from phidata.infra.k8s.resource.core.v1.namespace import Namespace, NamespaceSpec
from phidata.infra.k8s.create.common.labels import create_component_labels_dict
from phidata.infra.k8s.resource.meta.v1.object_meta import ObjectMeta
from phidata.utils.common import get_default_ns_name
from phidata.utils.log import logger


class CreateNamespace(BaseModel):
    ns: str
    app_name: str
    # Finalizers is an opaque list of values that must be empty to permanently remove object from storage.
    # More info: https://kubernetes.io/docs/tasks/administer-cluster/namespaces/
    finalizers: Optional[List[str]] = None
    labels: Optional[Dict[str, str]] = None

    def create(self) -> Optional[Namespace]:

        ns_name = self.ns if self.ns else get_default_ns_name(self.app_name)
        logger.debug(f"Init Namespace resource: {ns_name}")

        ns_labels = create_component_labels_dict(
            component_name=ns_name,
            app_name=self.app_name,
            labels=self.labels,
        )
        ns_spec = NamespaceSpec(finalizers=self.finalizers) if self.finalizers else None
        ns = Namespace(
            api_version=ApiVersion.CORE_V1,
            kind=Kind.NAMESPACE,
            metadata=ObjectMeta(
                name=ns_name,
                labels=ns_labels,
            ),
            spec=ns_spec,
        )
        return ns



from typing import List, Optional

from kubernetes.client import CoreV1Api
from kubernetes.client.models.v1_persistent_volume_claim import V1PersistentVolumeClaim
from kubernetes.client.models.v1_persistent_volume_claim_list import (
    V1PersistentVolumeClaimList,
)
from kubernetes.client.models.v1_persistent_volume_claim_spec import (
    V1PersistentVolumeClaimSpec,
)
from kubernetes.client.models.v1_status import V1Status
from pydantic import Field

from phidata.infra.k8s.api_client import K8sApiClient
from phidata.infra.k8s.enums.pv import PVAccessMode
from phidata.infra.k8s.resource.base import K8sResource, K8sObject
from phidata.infra.k8s.resource.core.v1.resource_requirements import (
    ResourceRequirements,
)
from phidata.utils.log import logger


class PersistentVolumeClaimSpec(K8sObject):
    """
    # https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#persistentvolumeclaim-v1-core
    """

    resource_type: str = "PersistentVolumeClaimSpec"

    access_modes: List[PVAccessMode] = Field(..., alias="accessModes")
    resources: ResourceRequirements
    storage_class_name: str = Field(..., alias="storageClassName")

    def get_k8s_object(
        self,
    ) -> V1PersistentVolumeClaimSpec:

        # Return a V1PersistentVolumeClaimSpec object
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume_claim_spec.py
        _v1_persistent_volume_claim_spec = V1PersistentVolumeClaimSpec(
            access_modes=self.access_modes,
            resources=self.resources.get_k8s_object(),
            storage_class_name=self.storage_class_name,
        )
        return _v1_persistent_volume_claim_spec


class PersistentVolumeClaim(K8sResource):
    """
    A PersistentVolumeClaim (PVC) is a request for storage by a user.
    It is similar to a Pod. Pods consume node resources and PVCs consume PV resources.
    A PersistentVolume (PV) is a piece of storage in the cluster that has been provisioned
    by an administrator or dynamically provisioned using Storage Classes.
    With Pak8, we prefer to use Storage Classes, read more about Dynamic provisioning here: https://kubernetes.io/docs/concepts/storage/persistent-volumes/#dynamic

    References:
        * Docs:
            https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.23/#persistentvolumeclaim-v1-core
            https://kubernetes.io/docs/concepts/storage/persistent-volumes/
        * Type: https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume_claim.py
    """

    resource_type: str = "PersistentVolumeClaim"

    spec: PersistentVolumeClaimSpec

    # List of attributes to include in the K8s manifest
    fields_for_k8s_manifest: List[str] = ["spec"]

    def get_k8s_object(self) -> V1PersistentVolumeClaim:
        """Creates a body for this PVC"""

        # Return a V1PersistentVolumeClaim object to create a ClusterRole
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_persistent_volume_claim.py
        _v1_persistent_volume_claim = V1PersistentVolumeClaim(
            api_version=self.api_version,
            kind=self.kind,
            metadata=self.metadata.get_k8s_object(),
            spec=self.spec.get_k8s_object(),
        )
        return _v1_persistent_volume_claim

    @staticmethod
    def get_from_cluster(
        k8s_client: K8sApiClient, namespace: Optional[str] = None, **kwargs
    ) -> Optional[List[V1PersistentVolumeClaim]]:
        """Reads PVCs from K8s cluster.

        Args:
            k8s_client: K8sApiClient for the cluster
            namespace: Namespace to use.
        """
        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pvc_list: Optional[V1PersistentVolumeClaimList] = None
        if namespace:
            logger.debug(f"Getting PVCs for ns: {namespace}")
            pvc_list = core_v1_api.list_namespaced_persistent_volume_claim(
                namespace=namespace, **kwargs
            )
        else:
            logger.debug("Getting PVCs for all namespaces")
            pvc_list = core_v1_api.list_persistent_volume_claim_for_all_namespaces(
                **kwargs
            )

        pvcs: Optional[List[V1PersistentVolumeClaim]] = None
        if pvc_list:
            pvcs = pvc_list.items
        logger.debug(f"pvcs: {pvcs}")
        logger.debug(f"pvcs type: {type(pvcs)}")
        return pvcs

    def _create(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        k8s_object: V1PersistentVolumeClaim = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Creating: {}".format(self.get_resource_name()))
        v1_persistent_volume_claim: V1PersistentVolumeClaim = (
            core_v1_api.create_namespaced_persistent_volume_claim(
                namespace=namespace, body=k8s_object
            )
        )
        # logger.debug("Created: {}".format(v1_persistent_volume_claim))
        if v1_persistent_volume_claim.metadata.creation_timestamp is not None:
            logger.debug("PVC Created")
            self.active_resource = v1_persistent_volume_claim
            self.active_resource_class = V1PersistentVolumeClaim
            return True
        logger.error("PVC could not be created")
        return False

    def _read(self, k8s_client: K8sApiClient) -> Optional[V1PersistentVolumeClaim]:
        """Returns the "Active" PVC from the cluster"""

        namespace = self.get_namespace()
        active_pvc: Optional[V1PersistentVolumeClaim] = None
        active_pvcs: Optional[List[V1PersistentVolumeClaim]] = self.get_from_cluster(
            k8s_client=k8s_client,
            namespace=namespace,
        )
        # logger.debug(f"active_pvcs: {active_pvcs}")
        if active_pvcs is None:
            return None

        _active_pvcs_dict = {_pvc.metadata.name: _pvc for _pvc in active_pvcs}

        pvc_name = self.get_resource_name()
        if pvc_name in _active_pvcs_dict:
            active_pvc = _active_pvcs_dict[pvc_name]
            self.active_resource = active_pvc
            self.active_resource_class = V1PersistentVolumeClaim
            # logger.debug(f"Found {pvc_name}")
        return active_pvc

    def _update(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pvc_name = self.get_resource_name()
        k8s_object: V1PersistentVolumeClaim = self.get_k8s_object()
        namespace = self.get_namespace()

        logger.debug("Updating: {}".format(pvc_name))
        v1_persistent_volume_claim: V1PersistentVolumeClaim = (
            core_v1_api.patch_namespaced_persistent_volume_claim(
                name=pvc_name, namespace=namespace, body=k8s_object
            )
        )
        # logger.debug("Updated:\n{}".format(pformat(v1_persistent_volume_claim.to_dict(), indent=2)))
        if v1_persistent_volume_claim.metadata.creation_timestamp is not None:
            logger.debug("PVC Updated")
            self.active_resource = v1_persistent_volume_claim
            self.active_resource_class = V1PersistentVolumeClaim
            return True
        logger.error("PVC could not be updated")
        return False

    def _delete(self, k8s_client: K8sApiClient) -> bool:

        core_v1_api: CoreV1Api = k8s_client.core_v1_api
        pvc_name = self.get_resource_name()
        namespace = self.get_namespace()

        logger.debug("Deleting: {}".format(pvc_name))
        self.active_resource = None
        # https://github.com/kubernetes-client/python/blob/master/kubernetes/client/models/v1_status.py
        _delete_status: V1Status = (
            core_v1_api.delete_namespaced_persistent_volume_claim(
                name=pvc_name, namespace=namespace
            )
        )
        # logger.debug("_delete_status: {}".format(pformat(_delete_status, indent=2)))
        if _delete_status.status == "Success":
            logger.debug("PVC Deleted")
            return True
        logger.error("PVC could not be deleted")
        return False



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, List, Union, Any
from typing_extensions import Literal

from phidata.app.db import DbApp
from phidata.app import PhidataApp, PhidataAppArgs
from phidata.constants import (
    SCRIPTS_DIR_ENV_VAR,
    STORAGE_DIR_ENV_VAR,
    META_DIR_ENV_VAR,
    PRODUCTS_DIR_ENV_VAR,
    NOTEBOOKS_DIR_ENV_VAR,
    WORKSPACE_CONFIG_DIR_ENV_VAR,
    PHIDATA_RUNTIME_ENV_VAR,
)
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    VolumeType,
    HostPathVolumeSource,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger

default_databox_name: str = "databox"


class DataboxArgs(PhidataAppArgs):
    name: str = default_databox_name
    version: str = "1"
    enabled: bool = True

    # Image args
    image_name: str = "phidata/databox"
    image_tag: str = "2.3.0"
    entrypoint: Optional[Union[str, List]] = None
    command: Optional[Union[str, List]] = None

    # Mount the workspace directory on the container
    mount_workspace: bool = True
    workspace_volume_name: str = "databox-ws-volume"
    # Path to mount the workspace volume under
    # This is the parent directory for the workspace on the container
    # i.e. the ws is mounted as a subdir in this dir
    # eg: if ws name is: idata, workspace path would be: /usr/local/workspaces/idata
    workspace_parent_container_path: str = "/usr/local/workspaces"
    # NOTE: On DockerContainers the workspace_root_path is mounted to workspace_dir
    # because we assume that DockerContainers are running locally on the user's machine
    # On K8sContainers, we usually load the workspace_dir from git using a git-sync sidecar container
    create_git_sync_sidecar: bool = True
    git_sync_repo: Optional[str] = None
    git_sync_branch: Optional[str] = None
    git_sync_wait: int = 1
    # But when running k8s locally, we can mount the workspace using
    # host path as well.
    k8s_mount_local_workspace = False

    # Install python dependencies using a requirements.txt file
    install_requirements: bool = False
    # Path to the requirements.txt file relative to the workspace_root
    requirements_file_path: str = "requirements.txt"

    # Mount aws config on the container
    # Only on DockerContainers, for K8sContainers use IamRole
    mount_aws_config: bool = False
    # Aws config dir on the host
    aws_config_path: Path = Path.home().resolve().joinpath(".aws")
    # Aws config dir on the container
    aws_config_container_path: str = "/root/.aws"

    # Only on DockerContainers
    # Mount airflow home from container to host machine
    # Useful when debugging the airflow conf
    mount_airflow_home: bool = False
    # Path to the dir on host machine relative to the workspace root
    airflow_home_dir: str = "airflow"
    # Path to airflow home on the container
    airflow_home_container_path: str = "/usr/local/airflow"

    # Configure airflow
    # If init_airflow = True, this databox initializes airflow and
    # sets the env var INIT_AIRFLOW = True
    # INIT_AIRFLOW = True is required by phidata to build dags
    init_airflow: bool = True
    # The AIRFLOW_ENV defines the current airflow runtime and can be used by DAGs to separate dev vs prd code
    airflow_env: Literal["dev", "stg", "prd"] = "dev"
    # If use_products_as_airflow_dags = True
    # set the AIRFLOW__CORE__DAGS_FOLDER to the products_dir
    use_products_as_airflow_dags: bool = True
    # If use_products_as_airflow_dags = False
    # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_path
    # airflow_dags_path is the directory in the container containing the airflow dags
    airflow_dags_path: Optional[str] = None
    # Creates an airflow admin with username: admin, pass: admin
    create_airflow_admin_user: bool = False
    airflow_executor: Literal[
        "DebugExecutor",
        "LocalExecutor",
        "SequentialExecutor",
        "CeleryExecutor",
        "CeleryKubernetesExecutor",
        "DaskExecutor",
        "KubernetesExecutor",
    ] = "SequentialExecutor"

    # Configure airflow db
    # If True, initialize the airflow_db on this databox
    # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True
    #   Locally, airflow_db uses sqllite
    # If using the databox with an external Airflow db, set init_airflow_db = False
    init_airflow_db: Optional[bool] = None
    # Upgrade the airflow db
    upgrade_airflow_db: bool = False
    wait_for_airflow_db: bool = False
    # Connect to database using DbApp
    airflow_db_app: Optional[DbApp] = None
    # Connect to database manually
    airflow_db_user: Optional[str] = None
    airflow_db_password: Optional[str] = None
    airflow_db_schema: Optional[str] = None
    airflow_db_host: Optional[str] = None
    airflow_db_port: Optional[int] = None
    airflow_db_driver: str = "postgresql+psycopg2"
    # Airflow db connections in the format { conn_id: conn_url }
    # converted to env var: AIRFLOW_CONN__conn_id = conn_url
    db_connections: Optional[Dict] = None

    # Start airflow standalone
    start_airflow_standalone: bool = False
    # Open the airflow_standalone_container_port on the container
    # if start_airflow_standalone=True
    airflow_standalone_container_port: int = 8080
    # standalone port on the host machine
    airflow_standalone_host_port: int = 8080
    # standalone port name on K8sContainer
    airflow_standalone_port_name: str = "standalone"

    # Configure airflow scheduler
    # Init Airflow scheduler as a daemon process
    # DEPRECATED: use start_airflow_standalone
    init_airflow_scheduler: bool = False

    # Configure airflow webserver
    # Init Airflow webserver when the container starts
    # DEPRECATED: use start_airflow_standalone
    init_airflow_webserver: bool = False

    # Open the airflow_webserver_container_port on the container
    # if init_airflow_webserver=True
    # This is also used to set AIRFLOW__WEBSERVER__WEB_SERVER_PORT
    airflow_webserver_container_port: int = 7080
    # webserver port on the host machine
    airflow_webserver_host_port: int = 8080
    # webserver port name on K8sContainer
    airflow_webserver_port_name: str = "webserver"

    # Configure the container
    container_name: Optional[str] = None
    image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT
    # Only used by the DockerContainer
    container_detach: bool = True
    container_auto_remove: bool = True
    container_remove: bool = True
    # Overwrite the PYTHONPATH env var, which is usually set to the workspace_root_container_path
    python_path: Optional[str] = None
    # Add container labels
    container_labels: Optional[Dict[str, Any]] = None
    # NOTE: Available only for Docker
    # Add volumes to DockerContainer
    # container_volumes is a dictionary which configures the volumes to mount
    # inside the container. The key is either the host path or a volume name,
    # and the value is a dictionary with 2 keys:
    #   bind - The path to mount the volume inside the container
    #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
    # For example:
    # {
    #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
    #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
    # }
    container_volumes: Optional[Dict[str, dict]] = None

    # Add env variables to container env
    env: Optional[Dict[str, str]] = None
    # Read env variables from a file in yaml format
    env_file: Optional[Path] = None
    # Configure the ConfigMap used for env variables that are not Secret
    config_map_name: Optional[str] = None
    # Configure the Secret used for env variables that are Secret
    secret_name: Optional[str] = None
    # Read secrets from a file in yaml format
    secrets_file: Optional[Path] = None

    # Configure the databox deploy
    deploy_name: Optional[str] = None
    pod_name: Optional[str] = None
    replicas: int = 1
    pod_node_selector: Optional[Dict[str, str]] = None
    restart_policy: RestartPolicy = RestartPolicy.ALWAYS
    termination_grace_period_seconds: Optional[int] = None
    # Add deployment labels
    deploy_labels: Optional[Dict[str, Any]] = None

    # Other args
    load_examples: bool = False
    print_env_on_load: bool = True

    # Install phidata in development mode
    install_phidata_dev: bool = False
    phidata_volume_name: str = "databox-phidata-volume"
    phidata_dir_path: Path = Path.home().joinpath("lab", "phidata")
    phidata_dir_container_path: str = "/phidata"


class Databox(PhidataApp):
    def __init__(
        self,
        name: str = default_databox_name,
        version: str = "1",
        enabled: bool = True,
        # Image args,
        image_name: str = "phidata/databox",
        image_tag: str = "2.3.0",
        entrypoint: Optional[Union[str, List]] = None,
        command: Optional[Union[str, List]] = None,
        # Mount the workspace directory on the container,
        mount_workspace: bool = True,
        workspace_volume_name: str = "databox-ws-volume",
        # Path to mount the workspace volume under,
        # This is the parent directory for the workspace on the container,
        # i.e. the ws is mounted as a subdir in this dir,
        # eg: if ws name is: idata, workspace path would be: /usr/local/workspaces/idata,
        workspace_parent_container_path: str = "/usr/local/workspaces",
        # NOTE: On DockerContainers the workspace_root_path is mounted to workspace_dir,
        # because we assume that DockerContainers are running locally on the user's machine,
        # On K8sContainers, we load the workspace_dir from git using a git-sync sidecar container,
        create_git_sync_sidecar: bool = True,
        git_sync_repo: Optional[str] = None,
        git_sync_branch: Optional[str] = None,
        git_sync_wait: int = 1,
        # But when running k8s locally, we can mount the workspace using
        # host path as well.
        k8s_mount_local_workspace=False,
        # Install python dependencies using a requirements.txt file,
        install_requirements: bool = False,
        # Path to the requirements.txt file relative to the workspace_root
        requirements_file_path: str = "requirements.txt",
        # Only on DockerContainers, for K8sContainers use IamRole
        # Mount aws config on the container
        mount_aws_config: bool = False,
        # Aws config dir on the host,
        aws_config_path: Path = Path.home().resolve().joinpath(".aws"),
        # Aws config dir on the container,
        aws_config_container_path: str = "/root/.aws",
        # Mount aws config on the container
        # Only on DockerContainers, for K8sContainers use IamRole
        mount_airflow_home: bool = False,
        # Path to the dir on host machine relative to the workspace root,
        airflow_home_dir: str = "airflow",
        # Path to airflow home on the container,
        airflow_home_container_path: str = "/usr/local/airflow",
        # Configure airflow,
        # If init_airflow = True, this databox initializes airflow and
        # sets the env var INIT_AIRFLOW = True
        # INIT_AIRFLOW = True is required by phidata to build dags
        init_airflow: bool = True,
        # The AIRFLOW_ENV defines the current airflow runtime and can be used by DAGs to separate dev vs prd code
        airflow_env: Literal["dev", "stg", "prd"] = "dev",
        # If use_products_as_airflow_dags = True,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the products_dir,
        use_products_as_airflow_dags: bool = True,
        # If use_products_as_airflow_dags = False,
        # set the AIRFLOW__CORE__DAGS_FOLDER to the airflow_dags_path,
        # airflow_dags_path is the directory in the container containing the airflow dags,
        airflow_dags_path: Optional[str] = None,
        # Creates an airflow admin with username: admin, pass: admin,
        create_airflow_admin_user: bool = False,
        airflow_executor: Literal[
            "DebugExecutor",
            "LocalExecutor",
            "SequentialExecutor",
            "CeleryExecutor",
            "CeleryKubernetesExecutor",
            "DaskExecutor",
            "KubernetesExecutor",
        ] = "SequentialExecutor",
        # Configure airflow db,
        # If True, initialize the airflow_db on this databox,
        # If None, value is derived from init_airflow i.e. initialize the airflow_db if init_airflow = True,
        #   Locally, airflow_db uses sqllite,
        # If using the databox with an external Airflow db, set init_airflow_db = False,
        init_airflow_db: Optional[bool] = None,
        # Upgrade the airflow db
        upgrade_airflow_db: bool = False,
        wait_for_airflow_db: bool = False,
        # Connect to database using DbApp,
        airflow_db_app: Optional[DbApp] = None,
        # Connect to database manually,
        airflow_db_user: Optional[str] = None,
        airflow_db_password: Optional[str] = None,
        airflow_db_schema: Optional[str] = None,
        airflow_db_host: Optional[str] = None,
        airflow_db_port: Optional[str] = None,
        airflow_db_driver: str = "postgresql+psycopg2",
        # Airflow db connections in the format { conn_id: conn_url },
        # converted to env var: AIRFLOW_CONN__conn_id = conn_url,
        db_connections: Optional[Dict] = None,
        # Start airflow standalone
        start_airflow_standalone: bool = False,
        # Open the airflow_standalone_container_port on the container
        # if start_airflow_standalone=True
        airflow_standalone_container_port: int = 8080,
        # standalone port on the host machine
        airflow_standalone_host_port: int = 8080,
        # standalone port name on K8sContainer
        airflow_standalone_port_name: str = "standalone",
        # Configure airflow scheduler,
        # Init Airflow scheduler as a daemon process,
        # DEPRECATED: use start_airflow_standalone
        init_airflow_scheduler: bool = False,
        # Configure airflow webserver,
        # Init Airflow webserver when the container starts,
        # DEPRECATED: use start_airflow_standalone
        init_airflow_webserver: bool = False,
        # Open the airflow_webserver_container_port on the container
        # if init_airflow_webserver=True
        # This is also used to set AIRFLOW__WEBSERVER__WEB_SERVER_PORT
        airflow_webserver_container_port: int = 7080,
        # webserver port on the host machine,
        airflow_webserver_host_port: int = 8080,
        # webserver port name on K8sContainer,
        airflow_webserver_port_name: str = "webserver",
        # Configure the container,
        container_name: Optional[str] = None,
        image_pull_policy: ImagePullPolicy = ImagePullPolicy.IF_NOT_PRESENT,
        # Only used by the DockerContainer,
        container_detach: bool = True,
        container_auto_remove: bool = True,
        container_remove: bool = True,
        # Overwrite the PYTHONPATH env var, which is usually set to the workspace_root_container_path
        python_path: Optional[str] = None,
        # Add container labels
        container_labels: Optional[Dict[str, Any]] = None,
        # NOTE: Available only for Docker
        # Add volumes to DockerContainer
        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes: Optional[Dict[str, dict]] = None,
        # Add env variables to container env,
        env: Optional[Dict[str, str]] = None,
        # Read env variables from a file in yaml format,
        env_file: Optional[Path] = None,
        # Configure the ConfigMap used for env variables that are not Secret,
        config_map_name: Optional[str] = None,
        # Configure the Secret used for env variables that are Secret,
        secret_name: Optional[str] = None,
        # Read secrets from a file in yaml format,
        secrets_file: Optional[Path] = None,
        # Configure the databox deploy,
        deploy_name: Optional[str] = None,
        pod_name: Optional[str] = None,
        replicas: int = 1,
        pod_node_selector: Optional[Dict[str, str]] = None,
        restart_policy: RestartPolicy = RestartPolicy.ALWAYS,
        termination_grace_period_seconds: Optional[int] = None,
        # Add deployment labels
        deploy_labels: Optional[Dict[str, Any]] = None,
        # Other args,
        load_examples: bool = False,
        print_env_on_load: bool = True,
        # Additional args
        # If True, skip resource creation if active resources with the same name exist.
        use_cache: bool = True,
        install_phidata_dev: bool = False,
        phidata_volume_name: str = "devbox-phidata-volume",
        phidata_dir_path: Path = Path.home().joinpath("lab", "phidata"),
        phidata_dir_container_path: str = "/phidata",
    ):
        super().__init__()
        try:
            self.args: DataboxArgs = DataboxArgs(
                name=name,
                version=version,
                enabled=enabled,
                image_name=image_name,
                image_tag=image_tag,
                entrypoint=entrypoint,
                command=command,
                mount_workspace=mount_workspace,
                workspace_volume_name=workspace_volume_name,
                workspace_parent_container_path=workspace_parent_container_path,
                create_git_sync_sidecar=create_git_sync_sidecar,
                git_sync_repo=git_sync_repo,
                git_sync_branch=git_sync_branch,
                git_sync_wait=git_sync_wait,
                k8s_mount_local_workspace=k8s_mount_local_workspace,
                install_requirements=install_requirements,
                requirements_file_path=requirements_file_path,
                mount_aws_config=mount_aws_config,
                aws_config_path=aws_config_path,
                aws_config_container_path=aws_config_container_path,
                mount_airflow_home=mount_airflow_home,
                airflow_home_dir=airflow_home_dir,
                airflow_home_container_path=airflow_home_container_path,
                init_airflow=init_airflow,
                airflow_env=airflow_env,
                use_products_as_airflow_dags=use_products_as_airflow_dags,
                airflow_dags_path=airflow_dags_path,
                create_airflow_admin_user=create_airflow_admin_user,
                airflow_executor=airflow_executor,
                init_airflow_db=init_airflow_db,
                upgrade_airflow_db=upgrade_airflow_db,
                wait_for_airflow_db=wait_for_airflow_db,
                airflow_db_app=airflow_db_app,
                airflow_db_user=airflow_db_user,
                airflow_db_password=airflow_db_password,
                airflow_db_schema=airflow_db_schema,
                airflow_db_host=airflow_db_host,
                airflow_db_port=airflow_db_port,
                airflow_db_driver=airflow_db_driver,
                db_connections=db_connections,
                start_airflow_standalone=start_airflow_standalone,
                airflow_standalone_container_port=airflow_standalone_container_port,
                airflow_standalone_host_port=airflow_standalone_host_port,
                airflow_standalone_port_name=airflow_standalone_port_name,
                init_airflow_scheduler=init_airflow_scheduler,
                init_airflow_webserver=init_airflow_webserver,
                airflow_webserver_container_port=airflow_webserver_container_port,
                airflow_webserver_host_port=airflow_webserver_host_port,
                airflow_webserver_port_name=airflow_webserver_port_name,
                container_name=container_name,
                image_pull_policy=image_pull_policy,
                container_detach=container_detach,
                container_auto_remove=container_auto_remove,
                container_remove=container_remove,
                python_path=python_path,
                container_labels=container_labels,
                container_volumes=container_volumes,
                env=env,
                env_file=env_file,
                config_map_name=config_map_name,
                secret_name=secret_name,
                secrets_file=secrets_file,
                deploy_name=deploy_name,
                pod_name=pod_name,
                replicas=replicas,
                pod_node_selector=pod_node_selector,
                restart_policy=restart_policy,
                termination_grace_period_seconds=termination_grace_period_seconds,
                deploy_labels=deploy_labels,
                load_examples=load_examples,
                print_env_on_load=print_env_on_load,
                use_cache=use_cache,
                install_phidata_dev=install_phidata_dev,
                phidata_volume_name=phidata_volume_name,
                phidata_dir_path=phidata_dir_path,
                phidata_dir_container_path=phidata_dir_container_path,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def get_container_name(self) -> str:
        return self.args.container_name or get_default_container_name(self.args.name)

    def get_env_data_from_file(self) -> Optional[Dict[str, str]]:
        env_file_path = self.args.env_file
        if (
            env_file_path is not None
            and env_file_path.exists()
            and env_file_path.is_file()
        ):
            import yaml

            # logger.debug(f"Reading {env_file_path}")
            env_data_from_file = yaml.safe_load(env_file_path.read_text())
            if env_data_from_file is not None and isinstance(env_data_from_file, dict):
                return env_data_from_file
            else:
                print_error(f"Invalid env_file: {env_file_path}")
        return None

    def get_secret_data_from_file(self) -> Optional[Dict[str, str]]:
        secrets_file_path = self.args.secrets_file
        if (
            secrets_file_path is not None
            and secrets_file_path.exists()
            and secrets_file_path.is_file()
        ):
            import yaml

            # logger.debug(f"Reading {secrets_file_path}")
            secret_data_from_file = yaml.safe_load(secrets_file_path.read_text())
            if secret_data_from_file is not None and isinstance(
                secret_data_from_file, dict
            ):
                return secret_data_from_file
            else:
                print_error(f"Invalid secrets_file: {secrets_file_path}")
        return None

    ######################################################
    ## Docker Resources
    ######################################################

    def init_airflow_on_docker_container(self, container: DockerContainer) -> None:
        """
        Initialize airflow on a docker container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        products_dir_container_path = (
            workspace_root_container_path.joinpath(self.products_dir)
            if self.products_dir
            else None
        )

        # Airflow db connection
        airflow_db_user = self.args.airflow_db_user
        airflow_db_password = self.args.airflow_db_password
        airflow_db_schema = self.args.airflow_db_schema
        airflow_db_host = self.args.airflow_db_host
        airflow_db_port = self.args.airflow_db_port
        airflow_db_driver = self.args.airflow_db_driver
        if self.args.airflow_db_app is not None and isinstance(
            self.args.airflow_db_app, DbApp
        ):
            logger.debug(
                f"Reading db connection details from: {self.args.airflow_db_app.name}"
            )
            if airflow_db_user is None:
                airflow_db_user = self.args.airflow_db_app.get_db_user()
            if airflow_db_password is None:
                airflow_db_password = self.args.airflow_db_app.get_db_password()
            if airflow_db_schema is None:
                airflow_db_schema = self.args.airflow_db_app.get_db_schema()
            if airflow_db_host is None:
                airflow_db_host = self.args.airflow_db_app.get_db_host_docker()
            if airflow_db_port is None:
                airflow_db_port = self.args.airflow_db_app.get_db_port_docker()
            if airflow_db_driver is None:
                airflow_db_driver = self.args.airflow_db_app.get_db_driver()
        db_connection_url = f"{airflow_db_driver}://{airflow_db_user}:{airflow_db_password}@{airflow_db_host}:{airflow_db_port}/{airflow_db_schema}"

        airflow_env: Dict[str, str] = {
            "INIT_AIRFLOW": str(self.args.init_airflow),
            "AIRFLOW_ENV": self.args.airflow_env,
            "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
            "UPGRADE_AIRFLOW_DB": str(self.args.upgrade_airflow_db),
            "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
            "AIRFLOW_DB_USER": str(airflow_db_user),
            "AIRFLOW_DB_PASSWORD": str(airflow_db_password),
            "AIRFLOW_SCHEMA": str(airflow_db_schema),
            "AIRFLOW_DB_HOST": str(airflow_db_host),
            "AIRFLOW_DB_PORT": str(airflow_db_port),
            "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
            "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
            "INIT_AIRFLOW_STANDALONE": str(self.args.start_airflow_standalone),
            "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            "CREATE_AIRFLOW_ADMIN_USER": str(self.args.create_airflow_admin_user),
            "AIRFLOW__CORE__EXECUTOR": str(self.args.airflow_executor),
        }

        # Set the AIRFLOW__DATABASE__SQL_ALCHEMY_CONN
        if "None" not in db_connection_url:
            logger.debug(f"AIRFLOW__DATABASE__SQL_ALCHEMY_CONN: {db_connection_url}")
            airflow_env["AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"] = db_connection_url

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_products_as_airflow_dags:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = str(products_dir_container_path)
        elif self.args.airflow_dags_path is not None:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_path

        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue

        # if init_airflow_webserver = True
        # 1. Set the webserver port in the container env
        # 2. Open the airflow webserver port
        if self.args.init_airflow_webserver:
            # Set the webserver port in the container env
            airflow_env["AIRFLOW__WEBSERVER__WEB_SERVER_PORT"] = str(
                self.args.airflow_webserver_container_port
            )
            # Open the port
            ws_port: Dict[str, int] = {
                str(
                    self.args.airflow_webserver_container_port
                ): self.args.airflow_webserver_host_port,
            }
            if container.ports is None:
                container.ports = {}
            container.ports.update(ws_port)

        # if start_airflow_standalone = True
        # 1. Open the airflow standalone port
        if self.args.start_airflow_standalone:
            # Open the port
            standalone_port: Dict[str, int] = {
                str(
                    self.args.airflow_standalone_container_port
                ): self.args.airflow_standalone_host_port,
            }
            if container.ports is None:
                container.ports = {}
            container.ports.update(standalone_port)

        # Update the container.environment to include airflow_env
        if isinstance(container.environment, dict):
            container.environment.update(airflow_env)
        else:
            logger.warning(
                f"Could not update container environment because it is of type: {type(container.environment)}"
            )

    def get_docker_rg(
        self, docker_build_context: DockerBuildContext
    ) -> Optional[DockerResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} DockerResourceGroup")

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return None
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        requirements_file_container_path = workspace_root_container_path.joinpath(
            self.args.requirements_file_path
        )
        scripts_dir_container_path = (
            workspace_root_container_path.joinpath(self.scripts_dir)
            if self.scripts_dir
            else None
        )
        storage_dir_container_path = (
            workspace_root_container_path.joinpath(self.storage_dir)
            if self.storage_dir
            else None
        )
        meta_dir_container_path = (
            workspace_root_container_path.joinpath(self.meta_dir)
            if self.meta_dir
            else None
        )
        products_dir_container_path = (
            workspace_root_container_path.joinpath(self.products_dir)
            if self.products_dir
            else None
        )
        notebooks_dir_container_path = (
            workspace_root_container_path.joinpath(self.notebooks_dir)
            if self.notebooks_dir
            else None
        )
        workspace_config_dir_container_path = (
            workspace_root_container_path.joinpath(self.workspace_config_dir)
            if self.workspace_config_dir
            else None
        )

        # Container pythonpath
        python_path = self.args.python_path or str(workspace_root_container_path)

        # Container Environment
        container_env: Dict[str, str] = {
            # Env variables used by data workflows and data assets
            "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            "PYTHONPATH": python_path,
            PHIDATA_RUNTIME_ENV_VAR: "airflow",
            SCRIPTS_DIR_ENV_VAR: str(scripts_dir_container_path),
            STORAGE_DIR_ENV_VAR: str(storage_dir_container_path),
            META_DIR_ENV_VAR: str(meta_dir_container_path),
            PRODUCTS_DIR_ENV_VAR: str(products_dir_container_path),
            NOTEBOOKS_DIR_ENV_VAR: str(notebooks_dir_container_path),
            WORKSPACE_CONFIG_DIR_ENV_VAR: str(workspace_config_dir_container_path),
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": str(requirements_file_container_path),
            "MOUNT_WORKSPACE": str(self.args.mount_workspace),
            # Print env when the container starts
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
            "INSTALL_PHIDATA_DEV": str(self.args.install_phidata_dev),
            "PHIDATA_DIR_PATH": self.args.phidata_dir_container_path,
        }

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env using secrets_file
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env.update(secret_data_from_file)

        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Container Volumes
        # container_volumes is a dictionary which configures the volumes to mount
        # inside the container. The key is either the host path or a volume name,
        # and the value is a dictionary with 2 keys:
        #   bind - The path to mount the volume inside the container
        #   mode - Either rw to mount the volume read/write, or ro to mount it read-only.
        # For example:
        # {
        #   '/home/user1/': {'bind': '/mnt/vol2', 'mode': 'rw'},
        #   '/var/www': {'bind': '/mnt/vol1', 'mode': 'ro'}
        # }
        container_volumes = self.args.container_volumes or {}
        # Create a volume for the workspace dir
        if self.args.mount_workspace:
            workspace_root_path_str = str(self.workspace_root_path)
            workspace_root_container_path_str = str(workspace_root_container_path)
            logger.debug(f"Mounting: {workspace_root_path_str}")
            logger.debug(f"\tto: {workspace_root_container_path_str}")
            container_volumes[workspace_root_path_str] = {
                "bind": workspace_root_container_path_str,
                "mode": "rw",
            }
        # Create a volume for aws config
        if self.args.mount_aws_config:
            aws_config_path_str = str(self.args.aws_config_path)
            logger.debug(f"Mounting: {aws_config_path_str}")
            logger.debug(f"\tto: {self.args.aws_config_container_path}")
            container_volumes[aws_config_path_str] = {
                "bind": self.args.aws_config_container_path,
                "mode": "ro",
            }
            container_env[
                "AWS_CONFIG_FILE"
            ] = f"{self.args.aws_config_container_path}/config"
            container_env[
                "AWS_SHARED_CREDENTIALS_FILE"
            ] = f"{self.args.aws_config_container_path}/credentials"
        # Create a volume for airflow home
        if self.args.mount_airflow_home:
            airflow_home_path_str = str(
                self.workspace_root_path.joinpath(self.args.airflow_home_dir)
            )
            airflow_home_container_path_str = str(self.args.airflow_home_container_path)
            logger.debug(f"Mounting: {airflow_home_path_str}")
            logger.debug(f"\tto: {airflow_home_container_path_str}")
            container_volumes[airflow_home_path_str] = {
                "bind": airflow_home_container_path_str,
                "mode": "rw",
            }
        # Create a volume for phidata in dev mode
        if self.args.install_phidata_dev and self.args.phidata_dir_path is not None:
            phidata_dir_absolute_path_str = str(self.args.phidata_dir_path)
            logger.debug(f"Mounting: {phidata_dir_absolute_path_str}")
            logger.debug(f"\tto: {self.args.phidata_dir_container_path}")
            container_volumes[phidata_dir_absolute_path_str] = {
                "bind": self.args.phidata_dir_container_path,
                "mode": "rw",
            }

        # Create the container
        docker_container = DockerContainer(
            name=self.get_container_name(),
            image=get_image_str(self.args.image_name, self.args.image_tag),
            entrypoint=self.args.entrypoint,
            command=self.args.command,
            detach=self.args.container_detach,
            auto_remove=self.args.container_auto_remove,
            remove=self.args.container_remove,
            stdin_open=True,
            tty=True,
            labels=self.args.container_labels,
            environment=container_env,
            network=docker_build_context.network,
            volumes=container_volumes,
            use_cache=self.args.use_cache,
        )
        # Initialize airflow on container
        self.init_airflow_on_docker_container(docker_container)
        # logger.debug(f"Databox Container Env: {docker_container.environment}")

        docker_rg = DockerResourceGroup(
            name=app_name,
            enabled=self.args.enabled,
            network=DockerNetwork(name=docker_build_context.network),
            containers=[docker_container],
        )
        return docker_rg

    def init_docker_resource_groups(
        self, docker_build_context: DockerBuildContext
    ) -> None:
        docker_rg = self.get_docker_rg(docker_build_context)
        if docker_rg is not None:
            if self.docker_resource_groups is None:
                self.docker_resource_groups = OrderedDict()
            self.docker_resource_groups[docker_rg.name] = docker_rg

    ######################################################
    ## K8s Resources
    ######################################################

    def init_airflow_on_k8s_container(
        self, container: CreateContainer, k8s_resource_group: CreateK8sResourceGroup
    ) -> None:
        """
        Initialize airflow on a k8s container
        """

        if not self.args.init_airflow:
            return

        # Update init_airflow_db arg if None
        if self.args.init_airflow_db is None:
            self.args.init_airflow_db = True

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        products_dir_container_path = (
            workspace_root_container_path.joinpath(self.products_dir)
            if self.products_dir
            else None
        )

        # Airflow db connection
        airflow_db_user = self.args.airflow_db_user
        airflow_db_password = self.args.airflow_db_password
        airflow_db_schema = self.args.airflow_db_schema
        airflow_db_host = self.args.airflow_db_host
        airflow_db_port = self.args.airflow_db_port
        airflow_db_driver = self.args.airflow_db_driver
        if self.args.airflow_db_app is not None and isinstance(
            self.args.airflow_db_app, DbApp
        ):
            logger.debug(
                f"Reading db connection details from: {self.args.airflow_db_app.name}"
            )
            if airflow_db_user is None:
                airflow_db_user = self.args.airflow_db_app.get_db_user()
            if airflow_db_password is None:
                airflow_db_password = self.args.airflow_db_app.get_db_password()
            if airflow_db_schema is None:
                airflow_db_schema = self.args.airflow_db_app.get_db_schema()
            if airflow_db_host is None:
                airflow_db_host = self.args.airflow_db_app.get_db_host_k8s()
            if airflow_db_port is None:
                airflow_db_port = self.args.airflow_db_app.get_db_port_k8s()
            if airflow_db_driver is None:
                airflow_db_driver = self.args.airflow_db_app.get_db_driver()
        db_connection_url = f"{airflow_db_driver}://{airflow_db_user}:{airflow_db_password}@{airflow_db_host}:{airflow_db_port}/{airflow_db_schema}"

        airflow_env: Dict[str, str] = {
            "INIT_AIRFLOW": str(self.args.init_airflow),
            "AIRFLOW_ENV": self.args.airflow_env,
            "INIT_AIRFLOW_DB": str(self.args.init_airflow_db),
            "UPGRADE_AIRFLOW_DB": str(self.args.upgrade_airflow_db),
            "WAIT_FOR_AIRFLOW_DB": str(self.args.wait_for_airflow_db),
            "AIRFLOW_DB_USER": str(airflow_db_user),
            "AIRFLOW_DB_PASSWORD": str(airflow_db_password),
            "AIRFLOW_SCHEMA": str(airflow_db_schema),
            "AIRFLOW_DB_HOST": str(airflow_db_host),
            "AIRFLOW_DB_PORT": str(airflow_db_port),
            "INIT_AIRFLOW_SCHEDULER": str(self.args.init_airflow_scheduler),
            "INIT_AIRFLOW_WEBSERVER": str(self.args.init_airflow_webserver),
            "INIT_AIRFLOW_STANDALONE": str(self.args.start_airflow_standalone),
            "AIRFLOW__CORE__LOAD_EXAMPLES": str(self.args.load_examples),
            "CREATE_AIRFLOW_ADMIN_USER": str(self.args.create_airflow_admin_user),
            "AIRFLOW__CORE__EXECUTOR": str(self.args.airflow_executor),
        }

        # Set the AIRFLOW__DATABASE__SQL_ALCHEMY_CONN
        if "None" not in db_connection_url:
            logger.debug(f"AIRFLOW__DATABASE__SQL_ALCHEMY_CONN: {db_connection_url}")
            airflow_env["AIRFLOW__DATABASE__SQL_ALCHEMY_CONN"] = db_connection_url
            # TODO: use AIRFLOW__CORE__SQL_ALCHEMY_CONN for older versions

        # Set the AIRFLOW__CORE__DAGS_FOLDER
        if self.args.mount_workspace and self.args.use_products_as_airflow_dags:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = str(products_dir_container_path)
        elif self.args.airflow_dags_path is not None:
            airflow_env["AIRFLOW__CORE__DAGS_FOLDER"] = self.args.airflow_dags_path

        # Set the AIRFLOW__CONN_ variables
        if self.args.db_connections is not None:
            for conn_id, conn_url in self.args.db_connections.items():
                try:
                    af_conn_id = str("AIRFLOW_CONN_{}".format(conn_id)).upper()
                    airflow_env[af_conn_id] = conn_url
                except Exception as e:
                    logger.exception(e)
                    continue

        # if init_airflow_webserver = True
        # 1. Set the webserver port in the container env
        # 2. Open the airflow webserver port
        if self.args.init_airflow_webserver:
            # Set the webserver port in the container env
            airflow_env["AIRFLOW__WEBSERVER__WEB_SERVER_PORT"] = str(
                self.args.airflow_webserver_container_port
            )
            # Open the port
            ws_port = CreatePort(
                name=self.args.airflow_webserver_port_name,
                container_port=self.args.airflow_webserver_container_port,
            )
            if container.ports is None:
                container.ports = []
            container.ports.append(ws_port)

        # if start_airflow_standalone = True
        # 1. Open the airflow standalone port
        if self.args.start_airflow_standalone:
            # Open the port
            standalone_port = CreatePort(
                name=self.args.airflow_standalone_port_name,
                container_port=self.args.airflow_standalone_container_port,
            )
            if container.ports is None:
                container.ports = []
            container.ports.append(standalone_port)

        airflow_env_cm = CreateConfigMap(
            cm_name=get_default_configmap_name("databox-airflow"),
            app_name=self.args.name,
            data=airflow_env,
        )
        # Add airflow_env_cm to container env
        if container.envs_from_configmap is None:
            container.envs_from_configmap = []
        container.envs_from_configmap.append(airflow_env_cm.cm_name)
        if k8s_resource_group.config_maps is None:
            k8s_resource_group.config_maps = []
        k8s_resource_group.config_maps.append(airflow_env_cm)

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        app_name = self.args.name
        logger.debug(f"Building {app_name} K8sResourceGroup")

        # Define K8s resources
        config_maps: List[CreateConfigMap] = []
        secrets: List[CreateSecret] = []
        volumes: List[CreateVolume] = []
        containers: List[CreateContainer] = []

        # Workspace paths
        if self.workspace_root_path is None:
            logger.error("Invalid workspace_root_path")
            return None
        workspace_name = self.workspace_root_path.stem
        workspace_root_container_path = Path(
            self.args.workspace_parent_container_path
        ).joinpath(workspace_name)
        requirements_file_container_path = workspace_root_container_path.joinpath(
            self.args.requirements_file_path
        )
        scripts_dir_container_path = (
            workspace_root_container_path.joinpath(self.scripts_dir)
            if self.scripts_dir
            else None
        )
        storage_dir_container_path = (
            workspace_root_container_path.joinpath(self.storage_dir)
            if self.storage_dir
            else None
        )
        meta_dir_container_path = (
            workspace_root_container_path.joinpath(self.meta_dir)
            if self.meta_dir
            else None
        )
        products_dir_container_path = (
            workspace_root_container_path.joinpath(self.products_dir)
            if self.products_dir
            else None
        )
        notebooks_dir_container_path = (
            workspace_root_container_path.joinpath(self.notebooks_dir)
            if self.notebooks_dir
            else None
        )
        workspace_config_dir_container_path = (
            workspace_root_container_path.joinpath(self.workspace_config_dir)
            if self.workspace_config_dir
            else None
        )

        # Container pythonpath
        python_path = self.args.python_path or str(workspace_root_container_path)

        # Container Environment
        container_env: Dict[str, str] = {
            # Env variables used by data workflows and data assets
            "PHI_WORKSPACE_PARENT": str(self.args.workspace_parent_container_path),
            "PHI_WORKSPACE_ROOT": str(workspace_root_container_path),
            "PYTHONPATH": python_path,
            PHIDATA_RUNTIME_ENV_VAR: "airflow",
            SCRIPTS_DIR_ENV_VAR: str(scripts_dir_container_path),
            STORAGE_DIR_ENV_VAR: str(storage_dir_container_path),
            META_DIR_ENV_VAR: str(meta_dir_container_path),
            PRODUCTS_DIR_ENV_VAR: str(products_dir_container_path),
            NOTEBOOKS_DIR_ENV_VAR: str(notebooks_dir_container_path),
            WORKSPACE_CONFIG_DIR_ENV_VAR: str(workspace_config_dir_container_path),
            "INSTALL_REQUIREMENTS": str(self.args.install_requirements),
            "REQUIREMENTS_FILE_PATH": str(requirements_file_container_path),
            "MOUNT_WORKSPACE": str(self.args.mount_workspace),
            # Print env when the container starts
            "PRINT_ENV_ON_LOAD": str(self.args.print_env_on_load),
        }

        # Update the container env using env_file
        env_data_from_file = self.get_env_data_from_file()
        if env_data_from_file is not None:
            container_env.update(env_data_from_file)

        # Update the container env with user provided env
        if self.args.env is not None and isinstance(self.args.env, dict):
            container_env.update(self.args.env)

        # Create a ConfigMap to set the container env variables which are not Secret
        container_env_cm = CreateConfigMap(
            cm_name=self.args.config_map_name or get_default_configmap_name(app_name),
            app_name=app_name,
            data=container_env,
        )
        config_maps.append(container_env_cm)

        # Create a Secret to set the container env variables which are Secret
        secret_data_from_file = self.get_secret_data_from_file()
        if secret_data_from_file is not None:
            container_env_secret = CreateSecret(
                secret_name=self.args.secret_name or get_default_secret_name(app_name),
                app_name=app_name,
                string_data=secret_data_from_file,
            )
            secrets.append(container_env_secret)

        # If mount_workspace=True first check if the workspace
        # should be mounted locally, otherwise
        # Create a Sidecar git-sync container and volume
        if self.args.mount_workspace:
            if self.args.k8s_mount_local_workspace:
                workspace_root_path_str = str(self.workspace_root_path)
                workspace_root_container_path_str = str(workspace_root_container_path)
                logger.debug(f"Mounting: {workspace_root_path_str}")
                logger.debug(f"\tto: {workspace_root_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=self.args.workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_root_container_path_str,
                    volume_type=VolumeType.HOST_PATH,
                    host_path=HostPathVolumeSource(
                        path=workspace_root_path_str,
                    ),
                )
                volumes.append(workspace_volume)

            elif self.args.create_git_sync_sidecar:
                workspace_parent_container_path_str = str(
                    self.args.workspace_parent_container_path
                )
                logger.debug(f"Creating EmptyDir")
                logger.debug(f"\tat: {workspace_parent_container_path_str}")
                workspace_volume = CreateVolume(
                    volume_name=self.args.workspace_volume_name,
                    app_name=app_name,
                    mount_path=workspace_parent_container_path_str,
                    volume_type=VolumeType.EMPTY_DIR,
                )
                volumes.append(workspace_volume)

                if self.args.git_sync_repo is None:
                    print_error("git_sync_repo invalid")
                else:
                    git_sync_env = {
                        "GIT_SYNC_REPO": self.args.git_sync_repo,
                        "GIT_SYNC_ROOT": str(self.args.workspace_parent_container_path),
                        "GIT_SYNC_DEST": workspace_name,
                    }
                    if self.args.git_sync_branch is not None:
                        git_sync_env["GIT_SYNC_BRANCH"] = self.args.git_sync_branch
                    if self.args.git_sync_wait is not None:
                        git_sync_env["GIT_SYNC_WAIT"] = str(self.args.git_sync_wait)
                    git_sync_sidecar = CreateContainer(
                        container_name="git-sync-workspaces",
                        app_name=app_name,
                        image_name="k8s.gcr.io/git-sync",
                        image_tag="v3.1.1",
                        env=git_sync_env,
                        envs_from_configmap=[cm.cm_name for cm in config_maps]
                        if len(config_maps) > 0
                        else None,
                        envs_from_secret=[secret.secret_name for secret in secrets]
                        if len(secrets) > 0
                        else None,
                        volumes=[workspace_volume],
                    )
                    containers.append(git_sync_sidecar)

        container_labels: Optional[Dict[str, Any]] = self.args.container_labels
        if k8s_build_context.labels is not None:
            if container_labels:
                container_labels.update(k8s_build_context.labels)
            else:
                container_labels = k8s_build_context.labels
        # Create the container
        k8s_container = CreateContainer(
            container_name=self.get_container_name(),
            app_name=app_name,
            image_name=self.args.image_name,
            image_tag=self.args.image_tag,
            # Equivalent to docker images CMD
            args=[self.args.command]
            if isinstance(self.args.command, str)
            else self.args.command,
            # Equivalent to docker images ENTRYPOINT
            command=self.args.entrypoint,
            image_pull_policy=self.args.image_pull_policy,
            envs_from_configmap=[cm.cm_name for cm in config_maps]
            if len(config_maps) > 0
            else None,
            envs_from_secret=[secret.secret_name for secret in secrets]
            if len(secrets) > 0
            else None,
            volumes=volumes if len(volumes) > 0 else None,
            labels=container_labels,
        )
        containers.append(k8s_container)

        # Set default container for kubectl commands
        # https://kubernetes.io/docs/reference/labels-annotations-taints/#kubectl-kubernetes-io-default-container
        pod_annotations = {
            "kubectl.kubernetes.io/default-container": k8s_container.container_name
        }

        deploy_labels: Optional[Dict[str, Any]] = self.args.deploy_labels
        if k8s_build_context.labels is not None:
            if deploy_labels:
                deploy_labels.update(k8s_build_context.labels)
            else:
                deploy_labels = k8s_build_context.labels
        # Create the deployment
        k8s_deployment = CreateDeployment(
            deploy_name=self.args.deploy_name or get_default_deploy_name(app_name),
            pod_name=self.args.pod_name or get_default_pod_name(app_name),
            app_name=app_name,
            namespace=k8s_build_context.namespace,
            service_account_name=k8s_build_context.service_account_name,
            replicas=self.args.replicas,
            containers=containers if len(containers) > 0 else None,
            pod_node_selector=self.args.pod_node_selector,
            restart_policy=self.args.restart_policy,
            termination_grace_period_seconds=self.args.termination_grace_period_seconds,
            volumes=volumes if len(volumes) > 0 else None,
            labels=deploy_labels,
            pod_annotations=pod_annotations,
        )

        # Create the K8sResourceGroup
        k8s_resource_group = CreateK8sResourceGroup(
            name=self.args.name,
            enabled=self.args.enabled,
            config_maps=config_maps if len(config_maps) > 0 else None,
            secrets=secrets if len(secrets) > 0 else None,
            deployments=[k8s_deployment],
        )

        # Initialize airflow on container
        self.init_airflow_on_k8s_container(k8s_container, k8s_resource_group)

        return k8s_resource_group.create()

    def init_k8s_resource_groups(self, k8s_build_context: K8sBuildContext) -> None:
        k8s_rg = self.get_k8s_rg(k8s_build_context)
        if k8s_rg is not None:
            if self.k8s_resource_groups is None:
                self.k8s_resource_groups = OrderedDict()
            self.k8s_resource_groups[k8s_rg.name] = k8s_rg



from collections import OrderedDict
from pathlib import Path
from typing import Optional, Dict, List, Union, Any, cast

from pydantic import AnyUrl, ValidationError

from phidata.app.db import DbApp
from phidata.app.k8s.app import K8sApp, K8sAppArgs
from phidata.exceptions.app import AppInvalidException
from phidata.constants import (
    SCRIPTS_DIR_ENV_VAR,
    STORAGE_DIR_ENV_VAR,
    META_DIR_ENV_VAR,
    PRODUCTS_DIR_ENV_VAR,
    NOTEBOOKS_DIR_ENV_VAR,
    WORKSPACE_CONFIG_DIR_ENV_VAR,
    PHIDATA_RUNTIME_ENV_VAR,
)
from phidata.infra.docker.resource.network import DockerNetwork
from phidata.infra.docker.resource.container import DockerContainer
from phidata.infra.docker.resource.group import (
    DockerResourceGroup,
    DockerBuildContext,
)
from phidata.infra.k8s.create.apps.v1.deployment import CreateDeployment
from phidata.infra.k8s.create.core.v1.secret import CreateSecret
from phidata.infra.k8s.create.core.v1.config_map import CreateConfigMap
from phidata.infra.k8s.create.core.v1.container import CreateContainer
from phidata.infra.k8s.create.core.v1.volume import (
    CreateVolume,
    VolumeType,
)
from phidata.infra.k8s.create.common.port import CreatePort
from phidata.infra.k8s.create.group import CreateK8sResourceGroup
from phidata.infra.k8s.enums.image_pull_policy import ImagePullPolicy
from phidata.infra.k8s.enums.restart_policy import RestartPolicy
from phidata.infra.k8s.resource.group import (
    K8sResourceGroup,
    K8sBuildContext,
)
from phidata.utils.common import (
    get_image_str,
    get_default_container_name,
    get_default_configmap_name,
    get_default_secret_name,
    get_default_deploy_name,
    get_default_pod_name,
)
from phidata.utils.cli_console import print_error
from phidata.utils.log import logger


class K8sUrlArgs(K8sAppArgs):
    url: str
    file_name: Optional[str] = None
    cache_dir: Union[str, Path] = "k8s"


class K8sUrl(K8sApp):
    def __init__(
        self,
        url: str,
        name: str,
        file_name: Optional[str] = None,
        cache_dir: Union[str, Path] = "k8s",
        version: str = "1",
        enabled: bool = True,
        cache_manifests: bool = True,
    ):
        super().__init__()

        try:
            self.args: K8sUrlArgs = K8sUrlArgs(
                url=url,
                name=name,
                file_name=file_name,
                cache_dir=cache_dir,
                version=version,
                enabled=enabled,
                cache_manifests=cache_manifests,
            )
        except Exception as e:
            logger.error(f"Args for {self.__class__.__name__} are not valid")
            raise

    def best_guess_file_name(self) -> str:

        from phidata.utils.dttm import get_today_utc_date_str

        fn: Optional[str] = self.args.file_name
        if fn is None:
            _url_path: str = str(self.args.url)
            if "/" in _url_path:
                fn = _url_path.split("/")[-1]
        if fn is None:
            fn = get_today_utc_date_str()
        logger.debug(f"Best guess file name: {fn}")
        return fn

    def get_download_file_path(self) -> Optional[Path]:
        """
        Returns the file path to download the URL to
        """

        file_dir: Optional[Path] = None
        if isinstance(self.args.cache_dir, str):
            workspace_config_dir = self.workspace_config_dir
            if workspace_config_dir:
                file_dir = (
                    Path(workspace_config_dir)
                    .joinpath(self.args.cache_dir)
                    .joinpath(self.args.name)
                    .resolve()
                )
        elif isinstance(self.args.cache_dir, Path):
            file_dir = self.args.cache_dir

        if file_dir is not None:
            file_name: str = self.best_guess_file_name()
            return file_dir.joinpath(file_name)

        return None

    def prepare_manifest_file(self) -> Path:
        """
        Downloads the URL if needed and returns a Path object
        pointing to the downloaded file

        Raises Exception if file cannot be prepared
        """

        # Validate URL
        url = self.args.url
        if url is None:
            raise AppInvalidException("Invalid URL")

        # Validate file_path is available
        download_file_path = self.get_download_file_path()
        if download_file_path is None:
            raise AppInvalidException("Download FilePath invalid")

        if (
            self.args.cache_manifests
            and download_file_path.exists()
            and download_file_path.is_file()
        ):
            logger.debug("File exists")
            return download_file_path

        # Download URL contents
        logger.debug("Downloading")
        logger.debug(f"  Url: {url}")
        logger.debug(f"  To : {download_file_path}")

        # Create the parent directory if it does not exist
        file_dir = download_file_path.parent
        if not file_dir.exists():
            file_dir.mkdir(parents=True, exist_ok=True)

        # Download URL to file
        try:
            import httpx

            response: httpx.Response = httpx.get(url=url)
            # logger.debug("response: {}".format(type(response)))
            # logger.debug("response: {}".format(str(response)))

            if response.is_error:
                logger.error("Could not read URL")
                logger.error(response.reason_phrase)
                response.raise_for_status()

            # Download file as text
            response_text: str = response.text
            if response_text is not None and response_text != "":
                download_file_path.write_text(response_text)
                logger.info(f"Url downloaded to {str(download_file_path)}")
                return download_file_path

            # Download file as bytes
            response_bytes: bytes = response.content
            if response_bytes is not None and response_bytes != b"":
                download_file_path.write_bytes(response_bytes)
                logger.info(f"Url downloaded to {str(download_file_path)}")
                return download_file_path
        except Exception as e:
            logger.error(f"Could not download url: {url}")
            logger.exception(e)
        raise AppInvalidException("Could not create manifest file")

    def get_k8s_rg(
        self, k8s_build_context: K8sBuildContext
    ) -> Optional[K8sResourceGroup]:

        # Prepare manifest file and set the args
        self.args.manifest_file = self.prepare_manifest_file()
        return super().get_k8s_rg(k8s_build_context=k8s_build_context)



from typing import Any, Optional

from phidata.constants import AWS_REGION_ENV_VAR, AWS_PROFILE_ENV_VAR
from phidata.infra.base.resource import InfraResource
from phidata.infra.aws.api_client import AwsApiClient
from phidata.utils.cli_console import print_info, print_error
from phidata.utils.log import logger


class AwsObject(InfraResource):
    """Base class for Aws Objects"""

    aws_region: Optional[str] = None
    aws_profile: Optional[str] = None

    aws_client: Optional[AwsApiClient] = None

    def get_aws_region(self) -> Optional[str]:
        # use value if provided
        if self.aws_region:
            return self.aws_region

        # get from env if not provided
        # logger.info(f"Loading {AWS_REGION_ENV_VAR} from env")
        import os

        aws_region_env = os.getenv(AWS_REGION_ENV_VAR)
        # logger.info(f"{AWS_REGION_ENV_VAR}: {aws_region_env}")
        if aws_region_env is not None:
            self.aws_region = aws_region_env
        return self.aws_region

    def get_aws_profile(self) -> Optional[str]:
        # use value if provided
        if self.aws_profile:
            return self.aws_profile

        # get from env if not provided
        # logger.info(f"Loading {AWS_PROFILE_ENV_VAR} from env")
        import os

        aws_profile_env = os.getenv(AWS_PROFILE_ENV_VAR)
        # logger.info(f"{AWS_PROFILE_ENV_VAR}: {aws_profile_env}")
        if aws_profile_env is not None:
            self.aws_profile = aws_profile_env
        return self.aws_profile

    def get_aws_client(self) -> AwsApiClient:
        if self.aws_client is not None:
            return self.aws_client
        self.aws_client = AwsApiClient(
            aws_region=self.get_aws_region(), aws_profile=self.get_aws_profile()
        )
        return self.aws_client


class AwsResource(AwsObject):
    """Base class for Aws Resources"""

    service_name: str
    service_client: Optional[Any] = None
    service_resource: Optional[Any] = None

    resource_available: bool = False
    resource_updated: bool = False
    resource_deleted: bool = False

    def is_valid(self) -> bool:
        # SubResources can use this function to add validation checks
        return True

    def get_service_client(self, aws_client: AwsApiClient):
        if self.service_client is None:
            self.service_client = aws_client.boto3_session.client(
                service_name=self.service_name, region_name=aws_client.aws_region
            )
        return self.service_client

    def get_service_resource(self, aws_client: AwsApiClient):
        if self.service_resource is None:
            self.service_resource = aws_client.boto3_session.resource(
                service_name=self.service_name, region_name=aws_client.aws_region
            )
        return self.service_resource

    def _create(self, aws_client: AwsApiClient) -> bool:
        logger.error(f"@_create method not defined for {self.__class__.__name__}")
        return False

    def create(self, aws_client: Optional[AwsApiClient] = None) -> bool:
        """Creates the resource on the Aws Cluster

        Args:
            aws_client: The AwsApiClient for the current Cluster
        """
        if not self.is_valid():
            return False
        # Skip resource creation if skip_create = True
        if self.skip_create:
            print_info(f"Skipping create: {self.get_resource_name()}")
            return True

        client: AwsApiClient = aws_client or self.get_aws_client()
        if self.use_cache and self.is_active(client):
            self.resource_available = True
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} already active."
            )
        else:
            self.resource_available = self._create(client)
        if self.resource_available:
            # print_info(
            #     f"Running post-create steps for {self.get_resource_type()}: {self.get_resource_name()}."
            # )
            return self.post_create(client)
        return self.resource_available

    def post_create(self, aws_client: AwsApiClient) -> bool:
        # return True because this function is not used for most resources
        return True

    def _read(self, aws_client: AwsApiClient) -> Any:
        logger.error(f"@_read method not defined for {self.__class__.__name__}")
        return False

    def read(self, aws_client: Optional[AwsApiClient] = None) -> Any:
        """Reads the resource from the Aws Cluster

        Args:
            aws_client: The AwsApiClient for the current Cluster
        """
        if not self.is_valid():
            return None
        if self.use_cache and self.active_resource is not None:
            return self.active_resource

        client: AwsApiClient = aws_client or self.get_aws_client()
        return self._read(client)

    def _update(self, aws_client: AwsApiClient) -> Any:
        logger.error(f"@_update method not defined for {self.__class__.__name__}")
        return False

    def update(self, aws_client: Optional[AwsApiClient] = None) -> bool:
        """Updates the resource on the Aws Cluster

        Args:
            aws_client: The AwsApiClient for the current Cluster
        """
        if not self.is_valid():
            return False
        # Skip resource creation if update = True
        if self.skip_update:
            print_info(f"Skipping update: {self.get_resource_name()}")
            return True

        client: AwsApiClient = aws_client or self.get_aws_client()
        if self.is_active(client):
            self.resource_updated = self._update(client)
        else:
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} does not exist."
            )
            return True
        if self.resource_updated:
            # print_info(
            #     f"Running post-update steps for {self.get_resource_type()}: {self.get_resource_name()}."
            # )
            return self.post_update(client)
        return self.resource_updated

    def post_update(self, aws_client: AwsApiClient) -> bool:
        # return True because this function is not used for most resources
        return True

    def _delete(self, aws_client: AwsApiClient) -> Any:
        logger.error(f"@_delete method not defined for {self.__class__.__name__}")
        return False

    def delete(self, aws_client: Optional[AwsApiClient] = None) -> bool:
        """Deletes the resource from the Aws Cluster

        Args:
            aws_client: The AwsApiClient for the current Cluster
        """
        if not self.is_valid():
            return False
        # Skip resource deletion if skip_delete = True
        if self.skip_delete:
            print_info(f"Skipping delete: {self.get_resource_name()}")
            return True

        client: AwsApiClient = aws_client or self.get_aws_client()
        if self.is_active(client):
            self.resource_deleted = self._delete(client)
        else:
            print_info(
                f"{self.get_resource_type()}: {self.get_resource_name()} does not exist."
            )
            return True
        if self.resource_deleted:
            # print_info(
            #     f"Running post-delete steps for {self.get_resource_type()}: {self.get_resource_name()}."
            # )
            return self.post_delete(client)
        return self.resource_deleted

    def post_delete(self, aws_client: AwsApiClient) -> bool:
        # return True because this function is not used for most resources
        return True

    def is_active(self, aws_client: AwsApiClient) -> bool:
        """Returns True if the resource is active on the Aws Cluster"""
        active_resource = self.read(aws_client=aws_client)
        if active_resource is not None:
            return True
        return False
